"""Base Aggregator classes.

Attributes:
    registry: A read-only dictionary of aggregator factory names to their implementation
        classes.
"""
from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Mapping as _RunTimeCheckableMapping
from enum import Enum, auto
import inspect
from types import MappingProxyType
from typing import (
    Any,
    ClassVar,
    Dict,
    Generic,
    Iterable,
    Mapping,
    Optional,
    Type,
    Union,
    overload,
)

import numpy as np
import pandas as pd

from bitfount.federated.exceptions import AggregatorError
from bitfount.federated.logging import _get_federated_logger
from bitfount.federated.roles import _RolesMixIn
from bitfount.federated.shim import _load_default_tensor_shim
from bitfount.federated.types import AggregatorType
from bitfount.types import (
    T_DTYPE,
    T_FIELDS_DICT,
    T_NESTED_FIELDS,
    _BaseSerializableObjectMixIn,
    _SerializedWeights,
    _Weights,
)

logger = _get_federated_logger(__name__)


class AggregationType(Enum):
    """Supported types for the aggregator."""

    TENSOR_DICT = auto()
    PANDAS_DATAFRAME = auto()
    NUMPY_ARRAY = auto()

    UNSUPPORTED = auto()


class _BaseAggregator(ABC):
    """Base Aggregator from which all other aggregators must inherit."""

    def __init__(self, **kwargs: Any):
        self._tensor_shim = _load_default_tensor_shim()
        # Dictionary of non numeric columns to values
        self.non_numeric_columns: Dict[str, np.ndarray] = {}

    @staticmethod
    def _get_type(obj: Iterable[Any]) -> AggregationType:
        """Determine the type of the supplied objects from the workers.

        :::note

        This is used to determine whether the object is a tensor, pandas dataframe or
        numpy array.

        :::
        """
        obj_iter = iter(obj)
        first = next(obj_iter)
        obj_type = type(first)

        # Check that these are the same in all others
        for other in obj_iter:
            if obj_type != type(other):
                raise AggregatorError(
                    f"Algorithm outputs are not consistent between workers: "
                    f"all types should match {obj_type}"
                )

        if obj_type == np.ndarray:
            logger.debug("Numpy array detected for aggregation.")
            return AggregationType.NUMPY_ARRAY
        elif obj_type == pd.DataFrame:
            logger.debug("Pandas dataframe detected for aggregation.")
            return AggregationType.PANDAS_DATAFRAME
        elif isinstance(first, _RunTimeCheckableMapping):
            logger.debug("Tensor-like object detected for aggregation.")
            return AggregationType.TENSOR_DICT

        return AggregationType.UNSUPPORTED

    def _set_non_numeric_columns(
        self, algorithm_outputs: Iterable[pd.DataFrame]
    ) -> None:
        """Extracts the non-numeric columns for an iterable of DataFrames.

        Raises:
            AggregatorError: If the non-numeric columns are not the same for all
                dataframes.
        """
        for output in algorithm_outputs:
            output_dict: Dict[str, np.ndarray] = {
                col: output[col].to_numpy() for col in output.columns
            }
            for col, values in output_dict.items():
                if (
                    col in self.non_numeric_columns
                    and (self.non_numeric_columns[col] != values).any()
                ):
                    raise AggregatorError(
                        f"Non-numeric columns must be the same. "
                        f"Column {col} has different values between workers."
                    )
                elif not np.issubdtype(values.dtype, np.number):
                    self.non_numeric_columns[col] = values


class _BaseModellerAggregator(_BaseAggregator, Generic[T_DTYPE], ABC):
    """Base modeller-side aggregator."""

    @overload
    def run(
        self,
        algorithm_outputs: Mapping[str, _SerializedWeights],
        tensor_dtype: Optional[T_DTYPE] = None,
        **kwargs: Any,
    ) -> _Weights:
        """Averages tensor dictionaries and returns the average."""
        ...

    @overload
    def run(
        self,
        algorithm_outputs: Mapping[str, pd.DataFrame],
        tensor_dtype: Optional[T_DTYPE] = None,
        **kwargs: Any,
    ) -> pd.DataFrame:
        """Averages pandas dataframes and returns the average."""
        ...

    @overload
    def run(
        self,
        algorithm_outputs: Mapping[str, np.ndarray],
        tensor_dtype: Optional[T_DTYPE] = None,
        **kwargs: Any,
    ) -> np.ndarray:
        """Averages numpy arrays and returns the average."""
        ...

    @abstractmethod
    def run(
        self,
        algorithm_outputs: Union[
            Mapping[str, _SerializedWeights],
            Mapping[str, np.ndarray],
            Mapping[str, pd.DataFrame],
        ],
        tensor_dtype: Optional[T_DTYPE] = None,
        **kwargs: Any,
    ) -> Union[_Weights, pd.DataFrame, np.ndarray]:
        """Averages algorithm outputs and returns the average."""


class _BaseWorkerAggregator(_BaseAggregator, ABC):
    """Base worker-side aggregator."""

    @overload
    async def run(
        self, algorithm_output: _Weights, **kwargs: Any
    ) -> _SerializedWeights:
        """Converts tensors to list of floats and returns them."""
        ...

    @overload
    async def run(self, algorithm_output: np.ndarray, **kwargs: Any) -> np.ndarray:
        """Doesn't do anything - leaves the type as it is."""
        ...

    @overload
    async def run(self, algorithm_output: pd.DataFrame, **kwargs: Any) -> pd.DataFrame:
        """Doesn't do anything - leaves the type as it is."""
        ...

    @abstractmethod
    async def run(
        self, algorithm_output: Union[_Weights, np.ndarray, pd.DataFrame], **kwargs: Any
    ) -> Union[_SerializedWeights, np.ndarray, pd.DataFrame]:
        """Runs the worker-side aggregator."""


# The mutable underlying dict that holds the registry information
_registry: Dict[str, Type[_BaseAggregatorFactory]] = {}
# The read-only version of the registry that is allowed to be imported
registry: Mapping[str, Type[_BaseAggregatorFactory]] = MappingProxyType(_registry)


class _AggregatorWorkerFactory(ABC):
    """Defines the base worker() factory method for aggregation."""

    @abstractmethod
    def worker(self, **kwargs: Any) -> _BaseWorkerAggregator:
        """Return worker side of the Aggregator."""
        pass


class _BaseAggregatorFactory(ABC, _BaseSerializableObjectMixIn, _RolesMixIn):
    """Base aggregator factory from which all others should inherit.

    Attributes:
        class_names: The name of the aggregator class.
    """

    fields_dict: ClassVar[T_FIELDS_DICT] = {}
    nested_fields: ClassVar[T_NESTED_FIELDS] = {}

    def __init__(self, **kwargs: Any) -> None:
        self.class_name = AggregatorType[type(self).__name__].value

    @classmethod
    def __init_subclass__(cls, **kwargs: Any):
        if not inspect.isabstract(cls):
            logger.debug(f"Adding {cls.__name__}: {cls} to Protocol registry")
            _registry[cls.__name__] = cls

    @abstractmethod
    def modeller(self, **kwargs: Any) -> _BaseModellerAggregator:
        """Return modeller side of the Aggregator."""
        raise NotImplementedError



"""Tests schema.py."""
import copy
from pathlib import Path
import re
from typing import Dict, List, MutableMapping, Optional, Union
from unittest.mock import Mock, PropertyMock, create_autospec

from _pytest.logging import LogCaptureFixture
from marshmallow import ValidationError
import numpy as np
import pandas as pd
from pandas.core.dtypes.common import pandas_dtype
from pandas.testing import assert_frame_equal
import pytest
from pytest import fixture
from pytest_mock import MockerFixture
import sqlalchemy
import yaml

import bitfount
from bitfount.data.datasources.base_source import BaseSource
from bitfount.data.datasources.database_source import DatabaseSource
from bitfount.data.datasources.dataframe_source import DataFrameSource
from bitfount.data.exceptions import DataSourceError
from bitfount.data.schema import (
    BitfountSchema,
    BitfountSchemaError,
    TableSchema,
    _TableSchemaMarshmallowMixIn,
)
from bitfount.data.types import (
    CategoricalRecord,
    ContinuousRecord,
    SemanticType,
    _SemanticTypeValue,
)
from bitfount.data.utils import DatabaseConnection, _generate_dtypes_hash, _hash_str
from bitfount.types import _Dtypes, _JSONDict
from tests.utils import PytestRequest
from tests.utils.helper import (
    TABLE_NAME,
    create_dataset,
    get_warning_logs,
    integration_test,
    unit_test,
)

SCHEMA_DUMP_PATH: str = "schema_dump_test.yaml"


@fixture
def data() -> pd.DataFrame:
    """Returns dataset."""
    return create_dataset()


@fixture(params=["dataframe", "db", "db_multi_table", "db_query"])
def datasource(
    request: PytestRequest,
    data: pd.DataFrame,
    db_session: sqlalchemy.engine.base.Engine,
) -> BaseSource:
    """Returns datasource."""
    datasource: BaseSource
    if request.param == "db":
        db_conn = DatabaseConnection(db_session, table_names=["dummy_data"])
        datasource = DatabaseSource(db_conn, seed=420)
    elif request.param == "db_multi_table":
        db_conn = DatabaseConnection(
            db_session, table_names=["dummy_data", "dummy_data_2"]
        )
        datasource = DatabaseSource(db_conn, seed=420)
    elif request.param == "db_query":
        db_conn = DatabaseConnection(db_session, query="SELECT * FROM dummy_data")
        datasource = DatabaseSource(db_conn, seed=420)
    elif request.param == "dataframe":
        datasource = DataFrameSource(data, seed=420)
    else:
        raise ValueError(
            "Supported Data Sources are : dataframe, db, db_multi_table and db_query."
        )

    return datasource


@fixture
def image_datasource() -> DataFrameSource:
    """Datasource with image_col for testing."""
    data = create_dataset(image=True)
    ds = DataFrameSource(data, image_col=["image"])
    ds.load_data()
    return ds


@fixture
def table_schema() -> TableSchema:
    """An EMPTY TableSchema object."""
    return TableSchema(name=TABLE_NAME)


@fixture
def schema() -> BitfountSchema:
    """An EMPTY BitfountSchema object."""
    return BitfountSchema()


@fixture
def table_schema_with_data(
    datasource: BaseSource, table_schema: TableSchema
) -> TableSchema:
    """Schema with datasource already set."""
    if isinstance(datasource, DatabaseSource) and (
        table_names := datasource.table_names
    ):
        table_schema = TableSchema(name=table_names[0])
    table_schema.add_datasource_features(datasource)
    return table_schema


@fixture
def schema_with_data(datasource: BaseSource, schema: BitfountSchema) -> BitfountSchema:
    """Schema with datasource already set."""
    if isinstance(datasource, DatabaseSource) and (
        table_names := datasource.table_names
    ):
        schema.add_datasource_tables(
            datasource,
            table_name=table_names[0],
            force_stypes={table_names[0]: {"categorical": ["Date"]}},
        )
    else:
        schema.add_datasource_tables(
            datasource,
            table_name=TABLE_NAME,
            force_stypes={TABLE_NAME: {"categorical": ["Date"]}},
        )
    return schema


@unit_test
class TestSchemaSerialization:
    """Tests we can save and load schema as yaml."""

    @fixture
    def empty_schema_hash(self) -> str:
        """The expected hash for a schema with no datasources."""
        empty_list: list = []
        return _hash_str(str(empty_list))

    @fixture
    def bitfount_version(self) -> str:
        """The bitfount version."""
        return bitfount.__version__

    @fixture
    def empty_schema_metadata(
        self, bitfount_version: str, empty_schema_hash: str
    ) -> Dict[str, str]:
        """The expected metadata for a schema with no datasources."""
        return {
            "bitfount_version": bitfount_version,
            "hash": empty_schema_hash,
            "schema_version": "1",
        }

    def test_schema_dump_load(
        self, datasource: BaseSource, schema: BitfountSchema, tmp_path: Path
    ) -> None:
        """Tests dumping and loading data schema."""
        schema.add_datasource_tables(datasource, table_name=TABLE_NAME)
        file_path = tmp_path / SCHEMA_DUMP_PATH
        schema.dump(file_path)
        loaded_schema = BitfountSchema.load_from_file(file_path)

        assert (
            yaml.safe_load(schema.dumps())["tables"]
            == yaml.safe_load(loaded_schema.dumps())["tables"]
        )

    def test_multi_table_schema_dump_load(
        self,
        mock_engine: Mock,
        mock_pandas_read_sql_query: None,
        mocker: MockerFixture,
        schema: BitfountSchema,
        tmp_path: Path,
    ) -> None:
        """Tests dumping and loading multi-table schema."""
        db_conn = DatabaseConnection(
            mock_engine,
            table_names=["dummy_data", "dummy_data_2"],
        )
        ds = DatabaseSource(db_conn, seed=420)
        mock_dtypes = create_dataset().dtypes.to_dict()
        mocker.patch.object(ds, "get_dtypes", return_value=mock_dtypes)
        ds._table_hashes.add(_generate_dtypes_hash(mock_dtypes))
        mocker.patch.object(
            ds,
            "get_values",
            side_effect=lambda x, table_name: {
                col: create_dataset()[col].unique() for col in x
            },
        )
        schema.add_datasource_tables(ds)
        assert schema.table_names == ["dummy_data", "dummy_data_2"]

        file_path = tmp_path / SCHEMA_DUMP_PATH
        schema.dump(file_path)
        loaded_schema = BitfountSchema.load_from_file(file_path)
        assert loaded_schema.table_names == ["dummy_data", "dummy_data_2"]
        assert (
            yaml.safe_load(schema.dumps())["tables"]
            == yaml.safe_load(loaded_schema.dumps())["tables"]
        )

    def test_dump_produces_metadata(
        self, bitfount_version: str, empty_schema_hash: str, schema: BitfountSchema
    ) -> None:
        """Tests that metadata is output in the schema dump."""
        dumped = schema.dumps()
        loaded_yaml = yaml.safe_load(dumped)
        assert loaded_yaml["metadata"] == {
            "hash": empty_schema_hash,
            "bitfount_version": bitfount_version,
            "schema_version": "1",
        }

    def test_load_works_without_metadata(
        self, datasource: BaseSource, schema: BitfountSchema, tmp_path: Path
    ) -> None:
        """Tests loading works if metadata tag not present."""
        schema.add_datasource_tables(datasource, table_name=TABLE_NAME)
        dumped = schema.dumps()
        # Remove metadata from dumped schema and save to file
        loaded_yaml = yaml.safe_load(dumped)
        del loaded_yaml["metadata"]
        assert "metadata" not in loaded_yaml
        file_path = tmp_path / "schema.yaml"
        with open(file_path, "w") as f:
            yaml.dump(loaded_yaml, f)

        # Load as BitfountSchema
        loaded_schema = BitfountSchema.load_from_file(file_path)

        # Check orig_hash not set (outcome of metadata loading)
        assert loaded_schema._orig_hash is None

    def test_schema_dump_load_sorts_keys_alphabetically(
        self, datasource: BaseSource, tmp_path: Path
    ) -> None:
        """Tests that key ordering is made alphabetical by dump and load."""
        schema = BitfountSchema()
        schema.add_datasource_tables(datasource, table_name=TABLE_NAME)
        file_path = tmp_path / SCHEMA_DUMP_PATH
        schema.dump(file_path)
        loaded_schema = BitfountSchema.load_from_file(file_path)

        for table in schema.tables:
            for stype in table.features:
                semantic_type = SemanticType(stype)
                original_feature_names = table.get_feature_names(semantic_type)
                loaded_feature_names = loaded_schema.get_feature_names(
                    table.name, semantic_type
                )
                assert sorted(original_feature_names) == loaded_feature_names

    def test_schema_raises_error_basesource_flag(
        self,
        datasource: BaseSource,
    ) -> None:
        """Tests schema raises error when BaseSource is not initialised."""
        del datasource._base_source_init
        schema = BitfountSchema()
        with pytest.raises(DataSourceError):
            schema.add_datasource_tables(datasource, table_name=TABLE_NAME)

    def test_schema_json_dump_has_correct_casing(
        self, empty_schema_metadata: Dict[str, str]
    ) -> None:
        """Tests camelCased output of JSON dump.

        We don't want to be changing the casing of fields that are from our dataset,
        i.e. feature names, categorical classes, but we do want python variable names
        to be camelCased so that they are consistent with the JSON format.
        """
        bf_schema = BitfountSchema()
        bf_schema.tables = [TableSchema(TABLE_NAME)]
        schema = bf_schema.tables[0]
        schema.features["continuous"] = {
            "snake_cased_name2": ContinuousRecord(
                "snake_cased_name2", pandas_dtype("float32")
            ),
            "camelCasedName2": ContinuousRecord(
                "camelCasedName2", pandas_dtype("float32")
            ),
        }

        record_with_encoder_one = CategoricalRecord(
            feature_name="snake_cased_name1", dtype=pd.StringDtype()
        )
        record_with_encoder_two = CategoricalRecord(
            feature_name="camelCasedName1", dtype=pd.StringDtype()
        )

        record_with_encoder_one.encoder.add_values(
            np.array(["these_are", "snake_cased"])
        )
        record_with_encoder_two.encoder.add_values(np.array(["theseAre", "camelCased"]))

        schema.features["categorical"] = {
            "snake_cased_name1": record_with_encoder_one,
            "camelCasedName1": record_with_encoder_two,
        }

        dumped_schema = bf_schema.dumps()
        expected_schema = yaml.dump(
            {
                "tables": [
                    {
                        "name": TABLE_NAME,
                        "description": None,
                        "features": [
                            {
                                "featureName": "camelCasedName1",
                                "description": None,
                                "dtype": "string",
                                "encoder": {
                                    "classes": {"camelCased": 0, "theseAre": 1}
                                },
                                "semanticType": "categorical",
                            },
                            {
                                "featureName": "camelCasedName2",
                                "description": None,
                                "dtype": "float32",
                                "semanticType": "continuous",
                            },
                            {
                                "featureName": "snake_cased_name1",
                                "description": None,
                                "dtype": "string",
                                "encoder": {
                                    "classes": {"snake_cased": 0, "these_are": 1}
                                },
                                "semanticType": "categorical",
                            },
                            {
                                "featureName": "snake_cased_name2",
                                "description": None,
                                "dtype": "float32",
                                "semanticType": "continuous",
                            },
                        ],
                    }
                ],
                "metadata": empty_schema_metadata,
            },
            sort_keys=False,
        )
        assert dumped_schema == expected_schema

    def test_schema_load_invalid_stype(
        self, datasource: BaseSource, tmp_path: Path
    ) -> None:
        """Tests that semantic types must be predefined in the Enum."""
        schema = BitfountSchema()
        schema.add_datasource_tables(datasource, table_name=TABLE_NAME)
        file_path = tmp_path / SCHEMA_DUMP_PATH
        schema.dump(file_path)

        # Read in the file
        with open(file_path, "r") as file:
            filedata = file.read()

        # Replace the target string
        filedata = filedata.replace("continuous", "not_real_stype")

        # Write the file out again
        with open(file_path, "w") as file:
            file.write(filedata)

        with pytest.raises(ValidationError, match="['Unknown field.']"):
            BitfountSchema.load_from_file(file_path)

    def test_schema_load_invalid_dtype(
        self, datasource: BaseSource, tmp_path: Path
    ) -> None:
        """Tests schema load fails if invalid dtype.

        Tests that a custom error message is reported when an invalid dtype is
        provided to a ContinuousRecord.
        """
        schema = BitfountSchema()
        schema.add_datasource_tables(datasource, table_name=TABLE_NAME)

        schema.tables[0].features["continuous"]["A"].dtype = "THIS_IS_NOT_A_VALID_DTYPE"

        file_path = tmp_path / SCHEMA_DUMP_PATH
        schema.dump(file_path)
        with pytest.raises(
            ValidationError,
            match="Continuous record `dtype` expected a valid np.dtype or a pandas"
            + " dtype but received: `THIS_IS_NOT_A_VALID_DTYPE`.",
        ):
            BitfountSchema.load_from_file(file_path)

    def test_schema_with_image_dump_load(
        self, image_datasource: DataFrameSource, schema: BitfountSchema, tmp_path: Path
    ) -> None:
        """Tests dumping and (re)loading a schema gives equal item."""
        file_path = tmp_path / SCHEMA_DUMP_PATH
        schema.add_datasource_tables(image_datasource, table_name=TABLE_NAME)
        schema.dump(file_path)
        loaded_schema = BitfountSchema.load_from_file(file_path)

        assert (
            yaml.safe_load(schema.dumps())["tables"]
            == yaml.safe_load(loaded_schema.dumps())["tables"]
        )

    def test_schema_split_features(self) -> None:
        """Tests the pre-load split_features function."""
        schema_json_dict = {
            "features": [
                {
                    "featureName": "TARGET",
                    "description": None,
                    "dtype": "int64",
                    "encoder": {"classes": {"0": 0, "1": 1}},
                    "semanticType": "categorical",
                },
                {
                    "featureName": "age",
                    "description": None,
                    "dtype": "float64",
                    "semanticType": "continuous",
                },
            ]
        }
        schema = _TableSchemaMarshmallowMixIn._Schema()

        # Check split_features splits features by stype
        schema_1 = schema.split_features(schema_json_dict)
        assert "categoricalFeatures" in schema_1.keys()
        assert "continuousFeatures" in schema_1.keys()

        # Check that on the second call, it just returns the data
        schema_after_2_loads = schema.split_features(schema_1)
        assert schema_after_2_loads == schema_1

        # Check that it raises error if no 'features' in the data.
        with pytest.raises(ValueError):
            schema.split_features({"data": "value"})


@unit_test
class TestTableSchema:
    """Tests TableSchema class."""

    def test_schema_add_datasource(
        self, datasource: BaseSource, table_schema: TableSchema
    ) -> None:
        """Checks ability to add a dataset."""
        if isinstance(datasource, DatabaseSource):
            if table_names := datasource.table_names:
                table_schema = TableSchema(name=table_names[0])
        table_schema.add_datasource_features(datasource)
        datasource.load_data()
        if not datasource.multi_table:
            assert len(table_schema.get_feature_names()) == len(datasource.data.columns)
        assert (
            "I" in table_schema.features["text"]
        )  # I's values are letters of alphabet
        assert (
            table_schema.features["categorical"]["M"].encoder.size == 2
        )  # M's values are boolean

    def test_schema_expand_dataframe_str(self, table_schema: TableSchema) -> None:
        """Checks we can add extend when columns have string types."""
        data1 = pd.DataFrame([["a", "b"], ["c", "d"]], columns=["A", "B"])
        data2 = pd.DataFrame([["a", "b"], ["c", "d"]], columns=["B", "C"])
        expect1 = pd.DataFrame(
            [["a", "b", "nan"], ["c", "d", "nan"]], columns=["A", "B", "C"]
        )
        expect2 = pd.DataFrame(
            [["a", "b", "nan"], ["c", "d", "nan"]], columns=["B", "C", "A"]
        )
        ds1 = DataFrameSource(data1)
        ds2 = DataFrameSource(data2)
        table_schema.add_datasource_features(ds1)
        table_schema.add_datasource_features(ds2)
        data1_exp = table_schema._expand_dataframe(data1)
        data2_exp = table_schema._expand_dataframe(data2)
        assert_frame_equal(data1_exp, expect1)
        assert_frame_equal(data2_exp, expect2)

    def test_schema_expand_dataframe_bool(self, table_schema: TableSchema) -> None:
        """Checks we can add extend when columns have bool types."""
        data1 = pd.DataFrame([[True, False], [True, False]], columns=["A", "B"])
        data2 = pd.DataFrame([[True, False], [True, False]], columns=["B", "C"])
        expect1 = pd.DataFrame(
            [[True, False, "nan"], [True, False, "nan"]],
            columns=["A", "B", "C"],
        )
        expect2 = pd.DataFrame(
            [[True, False, "nan"], [True, False, "nan"]],
            columns=["B", "C", "A"],
        )
        ds1 = DataFrameSource(data1)
        ds2 = DataFrameSource(data2)
        table_schema.add_datasource_features(ds1)
        table_schema.add_datasource_features(ds2)
        data1_exp = table_schema._expand_dataframe(data1)
        data2_exp = table_schema._expand_dataframe(data2)
        assert_frame_equal(data1_exp, expect1)
        assert_frame_equal(data2_exp, expect2)

    def test_schema_expand_dataframe_float(self, table_schema: TableSchema) -> None:
        """Checks we can add extend when columns have float types."""
        data1 = pd.DataFrame([[1.0, 2.0], [3.0, 4.0]], columns=["A", "B"])
        data2 = pd.DataFrame([[1.0, 2.0], [3.0, 4.0]], columns=["B", "C"])
        expect1 = pd.DataFrame([[1.0, 2.0, 0], [3.0, 4.0, 0]], columns=["A", "B", "C"])
        expect2 = pd.DataFrame([[1.0, 2.0, 0], [3.0, 4.0, 0]], columns=["B", "C", "A"])
        ds1 = DataFrameSource(data1)
        ds2 = DataFrameSource(data2)
        table_schema.add_datasource_features(ds1)
        table_schema.add_datasource_features(ds2)
        data1_exp = table_schema._expand_dataframe(data1)
        data2_exp = table_schema._expand_dataframe(data2)
        assert_frame_equal(data1_exp, expect1)
        assert_frame_equal(data2_exp, expect2)

    def test_schema_expand_dataframe_int(self, table_schema: TableSchema) -> None:
        """Checks we can add extend when columns have float types."""
        data1 = pd.DataFrame([[1, 2], [3, 4]], columns=["A", "B"])
        data2 = pd.DataFrame([[1, 2], [3, 4]], columns=["B", "C"])
        expect1 = pd.DataFrame([[1, 2, 0], [3, 4, 0]], columns=["A", "B", "C"])
        expect2 = pd.DataFrame([[1, 2, 0], [3, 4, 0]], columns=["B", "C", "A"])
        ds1 = DataFrameSource(data1)
        ds2 = DataFrameSource(data2)
        table_schema.add_datasource_features(ds1)
        table_schema.add_datasource_features(ds2)
        data1_exp = table_schema._expand_dataframe(data1)
        data2_exp = table_schema._expand_dataframe(data2)
        assert_frame_equal(data1_exp, expect1)
        assert_frame_equal(data2_exp, expect2)

    def test_schema_add_datasource_with_image(
        self, image_datasource: DataFrameSource, table_schema: TableSchema
    ) -> None:
        """Tests add_dataframe_features works correctly."""
        table_schema.add_datasource_features(
            image_datasource, force_stype={"image": ["image"]}
        )
        assert len(table_schema.get_feature_names()) == len(
            image_datasource.data.columns
        )
        assert table_schema.features["categorical"]["M"].encoder.size == 2
        assert list(table_schema.features["image"].keys()) == ["image"]
        image_feature = table_schema.features["image"]["image"]
        assert dict(image_feature.dimensions) == {(50, 50): 4000}
        assert "RGB" in image_feature.modes
        assert "L" in image_feature.modes
        assert dict(image_feature.formats) == {"PNG": 4000}

    def test_schema_add_datasource_without_force_stype_doesnt_contain_image_col(
        self,
        image_datasource: DataFrameSource,
        table_schema: TableSchema,
    ) -> None:
        """Tests that image col is not added to images."""
        table_schema.add_datasource_features(image_datasource)
        assert "image" not in table_schema.features

    def test_force_stypes(
        self, image_datasource: DataFrameSource, table_schema: TableSchema
    ) -> None:
        """Tests that all force_stypes are added to schema."""
        force_stype: MutableMapping[_SemanticTypeValue, List[str]]
        force_stype = {
            "categorical": ["M", "J"],
            "continuous": ["A"],
            "text": ["N"],
            "image": ["image"],
            "blah": ["B"],  # type: ignore[dict-item] # Reason: see below.
        }
        # Added the 'blah' dict key to check that if a value is given a semantic type
        # which we don't support, it defaults to the semantic value based
        # on dtype.
        table_schema.add_datasource_features(image_datasource, force_stype=force_stype)
        assert "M" in table_schema.features["categorical"]
        assert "J" in table_schema.features["categorical"]
        assert "A" in table_schema.features["continuous"]
        assert "N" in table_schema.features["text"]
        assert "B" in table_schema.features["continuous"]
        assert "image" in table_schema.features["image"]

    def test_stype_images(self, table_schema: TableSchema) -> None:
        """Tests multiple images added to schema."""
        data = create_dataset(classification=False, multiimage=True, img_size=2)
        ds = DataFrameSource(data[["image1", "image2", "TARGET"]])
        ds.load_data()
        force_stype: MutableMapping[_SemanticTypeValue, List[str]]
        force_stype = {"image": ["image1", "image2"], "text": ["TARGET"]}
        table_schema.add_datasource_features(ds, force_stype=force_stype)
        assert "image1" in table_schema.features["image"]
        assert "image2" in table_schema.features["image"]
        assert "TARGET" in table_schema.features["text"]

    def test_encode_dataframe(self, table_schema: TableSchema) -> None:
        """Tests schema encodes dataframe correctly."""
        data = pd.DataFrame([["a", "b"], ["c", "d"]], columns=["A", "B"])
        ds = DataFrameSource(data)
        table_schema.add_datasource_features(
            ds, force_stype={"categorical": ["A", "B"]}
        )
        expect = pd.DataFrame(
            [
                [
                    table_schema.features["categorical"]["A"].encoder.classes["a"],
                    table_schema.features["categorical"]["B"].encoder.classes["b"],
                ],
                [
                    table_schema.features["categorical"]["A"].encoder.classes["c"],
                    table_schema.features["categorical"]["B"].encoder.classes["d"],
                ],
            ],
            columns=["A", "B"],
        )
        data_exp = table_schema._encode_dataframe(data)
        assert_frame_equal(data_exp, expect)

    def test_reduce_dataframe(self, table_schema: TableSchema) -> None:
        """Tests reduce_dataframe removes appropriate columns."""
        data = pd.DataFrame([["a", "b"], ["c", "d"]], columns=["A", "B"])
        ds = DataFrameSource(data)
        ds.load_data()
        table_schema.add_datasource_features(ds)
        data2 = pd.DataFrame(
            [["a", "b", "c"], ["d", "e", "f"]], columns=["A", "B", "C"]
        )
        expect = pd.DataFrame([["a", "b"], ["d", "e"]], columns=["A", "B"])
        data_exp = table_schema._reduce_dataframe(data2)
        data_exp = data_exp.reindex(sorted(data_exp.columns), axis=1)
        assert_frame_equal(data_exp, expect)

    def test_apply_types(self, table_schema: TableSchema) -> None:
        """Tests apply_types changes types of columns appropriately."""
        data = pd.DataFrame([[1.1, 2.2], [3.3, 4.4]], columns=["A", "B"])
        ds = DataFrameSource(data)
        table_schema.add_datasource_features(ds)
        data2 = pd.DataFrame([[5, 6], [7, 8]], columns=["A", "B"])
        expect = pd.DataFrame(
            [[5.0, 6.0], [7.0, 8.0]], columns=["A", "B"]
        ).convert_dtypes(convert_integer=False)
        data_exp = table_schema._apply_types(data2)
        assert_frame_equal(data_exp, expect)

    def test_apply_types_w_keep_cols(self, table_schema: TableSchema) -> None:
        """Tests apply_types changes types of columns appropriately."""
        data = pd.DataFrame([[1.1, 2.2, 5.5], [3.3, 4.4, 6.6]], columns=["A", "B", "C"])
        ds = DataFrameSource(data)
        table_schema.add_datasource_features(ds)
        data2 = pd.DataFrame([[5, 6, 1], [7, 8, 9]], columns=["A", "B", "C"])
        expect = pd.DataFrame(
            [[5.0, 6.0, 1], [7.0, 8.0, 9]], columns=["A", "B", "C"]
        ).convert_dtypes(convert_integer=False)
        data_exp = table_schema._apply_types(data2, selected_cols=["A", "B"])
        assert_frame_equal(data_exp, expect)

    def test_categorical_feature_sizes_with_no_ignore_cols(
        self, table_schema: TableSchema
    ) -> None:
        """Schema returns all categorical sizes if no ignore_cols provided."""
        data = pd.DataFrame([[True, False], [False, True]], columns=["A", "B"])
        ds = DataFrameSource(data)
        table_schema.add_datasource_features(ds)
        assert table_schema.get_categorical_feature_sizes() == [2, 2]

    def test_categorical_feature_sizes_with_ignore_cols(
        self, table_schema: TableSchema
    ) -> None:
        """Schema returns correct categorical sizes when ignore_cols is categorical."""
        data = pd.DataFrame([[True, False], [False, True]], columns=["A", "B"])
        ds = DataFrameSource(data)
        table_schema.add_datasource_features(ds)
        assert table_schema.get_categorical_feature_sizes(ignore_cols="A") == [2]

    def test_get_categorical_feature_size_raises_value_error_when_there_are_no_categorical_features(  # noqa: B950
        self, table_schema: TableSchema
    ) -> None:
        """Tests that `get_categorical_feature_size` raises a ValueError appropriately.

        When there are no categorical features.
        """
        with pytest.raises(ValueError, match="No categorical features."):
            table_schema.get_categorical_feature_size("feature")

    def test_get_categorical_feature_size_raises_value_error_when_missing(
        self, table_schema: TableSchema
    ) -> None:
        """Tests that `get_categorical_feature_size` raises a ValueError appropriately.

        When it is missing from the list of categorical features.
        """
        table_schema.features["categorical"] = {
            "feature_1": Mock(),
            "feature_2": Mock(),
        }
        with pytest.raises(
            ValueError,
            match="missing_feature feature not found in categorical features.",
        ):
            table_schema.get_categorical_feature_size("missing_feature")

    def test_get_categorical_feature_with_list(self, table_schema: TableSchema) -> None:
        """Tests that `get_categorical_feature_size` works properly with a list.

        If a list of features is provided, we only return the size of the encoder for
        the first element.
        """
        table_schema.features["categorical"] = {
            "feature_1": Mock(encoder=Mock(size=5)),
            "feature_2": Mock(),
        }
        assert (
            table_schema.get_categorical_feature_size(["feature_1", "feature_2"]) == 5
        )

    @pytest.mark.parametrize(
        argnames="ignore_cols",
        argvalues=(
            pytest.param(None, id="no ignore_cols"),
            pytest.param("A", id="string ignore_cols"),
            pytest.param(["A"], id="list ignore_cols"),
        ),
    )
    def test_get_num_categorical(
        self,
        ignore_cols: Optional[Union[str, List[str]]],
        table_schema: TableSchema,
    ) -> None:
        """Tests get_num_categorical for different types of ignore_cols."""
        # Create schema with categorical data columns
        data = pd.DataFrame([[True, False], [False, True]], columns=["A", "B"])
        ds = DataFrameSource(data)
        ds.load_data()
        table_schema.add_datasource_features(ds)

        num_categorical = table_schema.get_num_categorical(ignore_cols)

        if ignore_cols:
            # Only "B"
            assert num_categorical == 1
        else:
            # Both "A" and "B"
            assert num_categorical == 2

    @pytest.mark.parametrize(
        argnames="ignore_cols",
        argvalues=(
            pytest.param(None, id="no ignore_cols"),
            pytest.param("A", id="string ignore_cols"),
            pytest.param(["A"], id="list ignore_cols"),
        ),
    )
    def test_get_num_continuous(
        self,
        ignore_cols: Optional[Union[str, List[str]]],
        table_schema: TableSchema,
    ) -> None:
        """Tests get_num_continuous for different types of ignore_cols."""
        # Create schema with continuous data columns
        data = pd.DataFrame([[1, 2], [3, 4]], columns=["A", "B"])
        ds = DataFrameSource(data)
        ds.load_data()
        table_schema.add_datasource_features(ds)

        num_continuous = table_schema.get_num_continuous(ignore_cols)

        if ignore_cols:
            # Only "B"
            assert num_continuous == 1
        else:
            # Both "A" and "B"
            assert num_continuous == 2

    def test__eq__passes_same_item(self, table_schema: TableSchema) -> None:
        """Test __eq__ passes when items are exact same."""
        assert table_schema == table_schema

    def test__eq__passes_same_features(
        self, table_schema_with_data: TableSchema
    ) -> None:
        """Test __eq__ passes when items have same features and types."""
        other_schema = copy.deepcopy(table_schema_with_data)

        # Check unique objects
        assert table_schema_with_data is not other_schema
        # Check still viewed as equal
        assert table_schema_with_data == other_schema

    def test__eq__fails_not_schema(self, table_schema: TableSchema) -> None:
        """Test __eq__ fails when other is not a TableSchema."""
        assert table_schema != object()

    def test__eq__fails_diff_feature_names(
        self, table_schema_with_data: TableSchema
    ) -> None:
        """Test __eq__ fails when feature names differ."""
        other_schema = copy.deepcopy(table_schema_with_data)

        # Remove a categorical feature
        cat_features = other_schema.features["categorical"]
        cat_features.pop(list(cat_features.keys())[0])

        assert table_schema_with_data != other_schema

    def test__eq__fails_diff_feature_dtypes(
        self, table_schema_with_data: TableSchema
    ) -> None:
        """Test __eq__ fails when feature dtypes differ."""
        other_schema = copy.deepcopy(table_schema_with_data)

        # Extract and replace dtype of a categorical feature
        cat_features = other_schema.features["categorical"]
        record = list(cat_features.values())[0]
        if record.dtype is not np.uint8:
            record.dtype = np.uint8
        else:
            record.dtype = np.uint16

        assert table_schema_with_data != other_schema

    def test__eq__fails_diff_feature_stypes(
        self, table_schema_with_data: TableSchema
    ) -> None:
        """Test __eq__ fails when feature stypes differ."""
        other_schema = copy.deepcopy(table_schema_with_data)

        # Extract and replace stype of a categorical feature
        cat_features = other_schema.features["categorical"]
        feature_name, record = list(cat_features.items())[0]
        # Create mock record with a different stype
        mock_record = create_autospec(record)
        type(mock_record).semantic_type = PropertyMock(return_value="diff_stype")
        # Set as feature
        cat_features[feature_name] = mock_record

        assert table_schema_with_data != other_schema


@unit_test
class TestTableSchemaSplittingByDtype:
    """Tests the dtype_based_stype_split static method on TableSchema."""

    # qlalchemy.sql.sqltypes.DATE

    @pytest.mark.parametrize(
        "data",
        [
            # csv / pandas
            pd.DataFrame([1.0, 2.0], columns=["A"], dtype=np.float32).dtypes.to_dict(),
            # database
            {"A": float},
        ],
    )
    def test_float_dtype(self, data: _Dtypes) -> None:
        """Tests that floats are given the correct semantic type."""
        types = TableSchema._dtype_based_stype_split(data)
        assert types == {SemanticType.CONTINUOUS: ["A"]}

    @pytest.mark.parametrize(
        "data",
        [
            # csv / pandas
            pd.DataFrame([1, 2], columns=["A"], dtype=np.int32).dtypes.to_dict(),
            # database
            {"A": int},
        ],
    )
    def test_int_dtype(self, data: _JSONDict) -> None:
        """Tests that integers are given the correct semantic type."""
        types = TableSchema._dtype_based_stype_split(data)
        assert types == {SemanticType.CONTINUOUS: ["A"]}

    @pytest.mark.parametrize(
        "data",
        [
            # csv / pandas
            pd.DataFrame(["1", "2"], columns=["A"]).convert_dtypes().dtypes.to_dict(),
            # database
            {"A": str},
        ],
    )
    def test_str_dtype(self, data: _Dtypes) -> None:
        """Tests that strings are given the correct semantic type."""
        types = TableSchema._dtype_based_stype_split(data)
        assert types == {SemanticType.TEXT: ["A"]}

    @pytest.mark.parametrize(
        "data",
        [
            # csv / pandas
            pd.DataFrame([True, False], columns=["A"]).dtypes.to_dict(),
            # database
            {"A": bool},
        ],
    )
    def test_bool_dtype(self, data: _Dtypes) -> None:
        """Tests that booleans are given the correct semantic type."""
        types = TableSchema._dtype_based_stype_split(data)
        assert types == {SemanticType.CATEGORICAL: ["A"]}

    def test_object_dtype(self) -> None:
        """Tests that objects are given the correct semantic type."""

        class ComplexObject:
            ...

        data = pd.DataFrame([ComplexObject(), ComplexObject()], columns=["A"])
        types = TableSchema._dtype_based_stype_split(data.dtypes.to_dict())
        assert types == {SemanticType.CATEGORICAL: ["A"]}


class TestBitfountSchema:
    """Tests for the BitfountSchema class."""

    @unit_test
    def test_schema_not_frozen_when_created(self, schema: BitfountSchema) -> None:
        """Test that a newly created schema is not frozen."""
        assert schema._frozen is False

    @unit_test
    def test_schema_freeze(self, schema: BitfountSchema) -> None:
        """Test that BitfountSchema.freeze() freezes."""
        schema.freeze()
        assert schema._frozen is True

    @unit_test
    def test_schema_unfreeze(self, schema: BitfountSchema) -> None:
        """Test that BitfountSchema.unfreeze() unfreezes."""
        schema.freeze()
        assert schema._frozen is True
        schema.unfreeze()
        assert schema._frozen is False

    @unit_test
    def test_datasource_cannot_be_added_when_frozen(
        self, datasource: BaseSource, schema: BitfountSchema
    ) -> None:
        """Tests that add_datasource_tables() fails when the schema is frozen."""
        schema.freeze()
        with pytest.raises(
            BitfountSchemaError,
            match=re.escape("This schema is frozen. No more datasources can be added."),
        ):
            schema.add_datasource_tables(datasource, table_name=TABLE_NAME)

    @unit_test
    def test_hash(self, schema: BitfountSchema) -> None:
        """Tests that hash generation works."""
        schema._datasource_hashes = set(["world", "hello"])

        # Should sort the stored "hashes"
        assert schema.hash == _hash_str(str(sorted(set(["hello", "world"]))))

    @unit_test
    def test_freeze_raises_exception_if_hash_mismatch(
        self, schema: BitfountSchema
    ) -> None:
        """Tests that freeze raises an exception if previous hash doesn't match."""
        schema._orig_hash = "not_the_same_hash"
        with pytest.raises(
            BitfountSchemaError,
            match=re.escape(
                "This schema was generated against a different set of datasources "
                "and is incompatible with those selected. This may be due to "
                "changing column names or types. Please generate a new schema."
            ),
        ):
            schema.freeze()

    @unit_test
    def test_add_datasource_tables_requires_table_name_for_single_table_datasource(
        self, datasource: BaseSource, schema: BitfountSchema
    ) -> None:
        """Tests that add_datasource_tables raises a ValueError.

        If a single table datasource is passed, the table name must be specified.
        """
        if not datasource.multi_table:
            with pytest.raises(
                ValueError,
                match="Must provide a table name for single table datasources.",
            ):
                schema.add_datasource_tables(datasource)

    @integration_test
    def test_add_datasource_tables_with_multi_table_datasource(
        self,
        data: pd.DataFrame,
        db_session: sqlalchemy.engine.base.Engine,
        schema: BitfountSchema,
    ) -> None:
        """Tests that add_datasource_tables works with a multi-table datasource."""
        db_conn = DatabaseConnection(
            db_session, table_names=["dummy_data", "dummy_data_2"]
        )
        ds = DatabaseSource(db_conn, seed=420)
        schema.add_datasource_tables(
            ds,
            force_stypes={
                "dummy_data": {"text": ["Date"]},
                "dummy_data_2": {"text": ["Date"]},
            },
        )
        assert len(schema.tables) == 2
        assert schema.table_names == ["dummy_data", "dummy_data_2"]
        for table_schema in schema.tables:
            assert table_schema.features is not None
            assert len(table_schema.get_feature_names()) == len(data.columns)

    @unit_test
    def test_mock_add_datasource_tables_with_multi_table_datasource(
        self,
        data: pd.DataFrame,
        mock_engine: Mock,
        mock_pandas_read_sql_query: None,
        mocker: MockerFixture,
        schema: BitfountSchema,
    ) -> None:
        """Tests that add_datasource_tables works with a multi-table datasource."""
        db_conn = DatabaseConnection(
            mock_engine, table_names=["dummy_data", "dummy_data_2"]
        )
        ds = DatabaseSource(db_conn, seed=420)

        mock_dtypes = create_dataset().dtypes.to_dict()
        mocker.patch.object(ds, "get_dtypes", return_value=mock_dtypes)
        ds._table_hashes.add(_generate_dtypes_hash(mock_dtypes))

        mocker.patch.object(
            ds,
            "get_values",
            side_effect=lambda x, table_name: {
                col: create_dataset()[col].unique() for col in x
            },
        )

        schema.add_datasource_tables(ds)
        assert len(schema.tables) == 2
        assert schema.table_names == ["dummy_data", "dummy_data_2"]
        for table_schema in schema.tables:
            assert table_schema.features is not None
            assert len(table_schema.get_feature_names()) == len(data.columns)

    @unit_test
    def test_mock_add_datasource_tables_with_multi_table_datasource_table_name_warning(
        self,
        caplog: LogCaptureFixture,
        mock_engine: Mock,
        mock_pandas_read_sql_query: None,
        mocker: MockerFixture,
        schema: BitfountSchema,
    ) -> None:
        """Tests that add_datasource_tables works with a multi-table datasource."""
        db_conn = DatabaseConnection(
            mock_engine, table_names=["dummy_data", "dummy_data_2"]
        )
        ds = DatabaseSource(db_conn, seed=420)

        mock_dtypes = create_dataset().dtypes.to_dict()
        mocker.patch.object(ds, "get_dtypes", return_value=mock_dtypes)
        ds._table_hashes.add(_generate_dtypes_hash(mock_dtypes))

        mocker.patch.object(
            ds,
            "get_values",
            side_effect=lambda x, table_name: {
                col: create_dataset()[col].unique() for col in x
            },
        )

        schema.add_datasource_tables(ds, table_name="this_should_be_ignored")
        assert len(schema.tables) == 2
        assert schema.table_names == ["dummy_data", "dummy_data_2"]
        warning_logs = get_warning_logs(caplog)
        assert (
            "Ignoring table_name argument for multi-table datasource." in warning_logs
        )

    @unit_test
    def test_add_datasource_tables_adds_to_hash_list(
        self, datasource: BaseSource, mocker: MockerFixture, schema: BitfountSchema
    ) -> None:
        """Checks add_datasource_tables() adds to hash list.

        Each added datasource needs to be recorded, so we use the hash list to
        achieve this.
        """
        # Mock out datasource.hash
        mock_hash = PropertyMock(return_value="datasource_hash")
        mocker.patch.object(BaseSource, "hash", mock_hash)

        assert schema._datasource_hashes == set()

        schema.add_datasource_tables(datasource, table_name=TABLE_NAME)

        assert schema._datasource_hashes == set(["datasource_hash"])
        mock_hash.assert_called_once()

    @unit_test
    def test_get_table_schema(self, schema_with_data: BitfountSchema) -> None:
        """Tests that get_table_schema() returns the correct schema."""
        table_name = schema_with_data.table_names[0]
        assert (
            schema_with_data.get_table_schema(table_name) == schema_with_data.tables[0]
        )

    @unit_test
    def test_get_table_schema_raises_error_if_schema_not_found(
        self, schema_with_data: BitfountSchema
    ) -> None:
        """Tests that get_table_schema() raises an error if schema not found."""
        with pytest.raises(
            BitfountSchemaError, match="Table Not_a_table not found in schema."
        ):
            schema_with_data.get_table_schema("Not_a_table")

    @unit_test
    def test_get_feature_names(self, schema_with_data: BitfountSchema) -> None:
        """Test that get_feature_names is a wrapper around table schema."""
        table_name = schema_with_data.table_names[0]
        assert (
            schema_with_data.get_feature_names(table_name)
            == schema_with_data.tables[0].get_feature_names()
        )

    @unit_test
    def test_get_categorical_feature_sizes(
        self, schema_with_data: BitfountSchema
    ) -> None:
        """Test that get_categorical_feature_sizes is a wrapper around table schema."""
        table_name = schema_with_data.table_names[0]
        assert (
            schema_with_data.get_categorical_feature_sizes(table_name)
            == schema_with_data.tables[0].get_categorical_feature_sizes()
        )

    @unit_test
    def test_get_categorical_feature_size(
        self, schema_with_data: BitfountSchema
    ) -> None:
        """Test that get_categorical_feature_size is a wrapper around table schema."""
        table_name = schema_with_data.table_names[0]
        assert schema_with_data.get_categorical_feature_size(
            table_name, "Date"
        ) == schema_with_data.tables[0].get_categorical_feature_size("Date")

    @unit_test
    def test_empty_schema_cannot_be_applied(
        self, data: pd.DataFrame, schema: BitfountSchema
    ) -> None:
        """Tests that applying an empty schema raises an error."""
        with pytest.raises(
            BitfountSchemaError,
            match="No tables in schema.",
        ):
            schema.apply(data)

    @unit_test
    def test_multi_table_schema_cannot_be_applied_to_single_table_data(
        self, data: pd.DataFrame, schema_with_data: BitfountSchema
    ) -> None:
        """Tests that applying a multi-table schema to single table data errors."""
        schema_with_data.tables.append(schema_with_data.tables[0])
        with pytest.raises(
            BitfountSchemaError,
            match="Can't apply a multi-table schema to a single dataframe.",
        ):
            schema_with_data.apply(data)

    @unit_test
    def test_single_table_schema_apply_method_applies_schema(
        self, data: pd.DataFrame, schema: BitfountSchema
    ) -> None:
        """Tests that apply() applies the schema to the data."""
        mock_table_schema = Mock(spec=TableSchema)
        schema.tables = [mock_table_schema]
        schema.apply(data)
        mock_table_schema.apply.assert_called_once()

    @unit_test
    def test__eq__passes_same_item(self, schema: BitfountSchema) -> None:
        """Test __eq__ passes when items are exact same."""
        assert schema == schema

    @unit_test
    def test__eq__passes_same_features(self, schema_with_data: BitfountSchema) -> None:
        """Test __eq__ passes when items have same features and types."""
        other_schema = copy.deepcopy(schema_with_data)

        # Check unique objects
        assert schema_with_data is not other_schema
        # Check still viewed as equal
        assert schema_with_data == other_schema

    @unit_test
    def test__eq__fails_not_schema(self, schema: BitfountSchema) -> None:
        """Test __eq__ fails when other is not a BitfountSchema."""
        assert schema != object()

    @unit_test
    def test__eq__fails_diff_number_of_tables(
        self, schema_with_data: BitfountSchema
    ) -> None:
        """Test __eq__ fails when number of tables differ."""
        other_schema = copy.deepcopy(schema_with_data)
        other_schema.tables.append(other_schema.tables[0])

        assert schema_with_data != other_schema

    @unit_test
    def test__eq__fails_diff_table_names(
        self, schema_with_data: BitfountSchema
    ) -> None:
        """Test __eq__ fails when table names differ."""
        other_schema = copy.deepcopy(schema_with_data)
        other_schema.tables[0].name = "other_name"

        assert schema_with_data != other_schema

    @unit_test
    def test__eq__fails_diff_feature_names(
        self, schema_with_data: BitfountSchema
    ) -> None:
        """Test __eq__ fails when feature names differ."""
        other_schema = copy.deepcopy(schema_with_data)

        # Remove a categorical feature
        cat_features = other_schema.tables[0].features["categorical"]
        cat_features.pop(list(cat_features.keys())[0])

        assert schema_with_data != other_schema

    @unit_test
    def test__eq__fails_diff_feature_dtypes(
        self, schema_with_data: BitfountSchema
    ) -> None:
        """Test __eq__ fails when feature dtypes differ."""
        other_schema = copy.deepcopy(schema_with_data)

        # Extract and replace dtype of a categorical feature
        cat_features = other_schema.tables[0].features["categorical"]
        record = list(cat_features.values())[0]
        if record.dtype is not np.uint8:
            record.dtype = np.uint8
        else:
            record.dtype = np.uint16

        assert schema_with_data != other_schema

    @unit_test
    def test__eq__fails_diff_feature_stypes(
        self, schema_with_data: BitfountSchema
    ) -> None:
        """Test __eq__ fails when feature stypes differ."""
        other_schema = copy.deepcopy(schema_with_data)

        # Extract and replace stype of a categorical feature
        cat_features = other_schema.tables[0].features["categorical"]
        feature_name, record = list(cat_features.items())[0]
        # Create mock record with a different stype
        mock_record = create_autospec(record)
        type(mock_record).semantic_type = PropertyMock(return_value="diff_stype")
        # Set as feature
        cat_features[feature_name] = mock_record

        assert schema_with_data != other_schema



"""Contains PyTorch implementations of the BitfountModel paradigm."""
from abc import abstractmethod
from collections import defaultdict
import logging
import os
from typing import Any, Dict, List, MutableMapping, Optional, Tuple, Union, cast

import numpy as np
import pytorch_lightning as pl
from pytorch_lightning.callbacks import Callback
from pytorch_lightning.callbacks.progress import TQDMProgressBar
import torch
from torch import nn as nn
from torch.optim.lr_scheduler import _LRScheduler
from torch.utils.data import DataLoader as PyTorchDataLoader

from bitfount.backends.pytorch.data.dataloaders import _BasePyTorchBitfountDataLoader
from bitfount.backends.pytorch.federated.mixins import _PyTorchDistributedModelMixIn
from bitfount.backends.pytorch.models.base_models import (
    _STEP_OUTPUT,
    _TORCH_DTYPES,
    _OptimizerType,
)
from bitfount.backends.pytorch.utils import _autodetect_gpu
from bitfount.data.databunch import BitfountDataBunch
from bitfount.data.dataloaders import BitfountDataLoader
from bitfount.data.datasources.base_source import BaseSource
from bitfount.metrics import Metric
from bitfount.models.base_models import ClassifierMixIn, ModelContext
from bitfount.models.bitfount_model import BitfountModel
from bitfount.types import T_DTYPE, _StrAnyDict
from bitfount.utils import _merge_list_of_dicts, delegates

logger = logging.getLogger(__name__)


@delegates()
class PyTorchBitfountModel(
    _PyTorchDistributedModelMixIn, BitfountModel, pl.LightningModule
):
    """Blueprint for a pytorch custom model in the lightning format.

    This class must be subclassed in its own module. A `Path` to the module containing
    the subclass can then be passed to `BitfountModelReference` and on to your
    `Algorithm` of choice which will send the model to Bitfount Hub.

    To get started, just implement the abstract methods in this class. For more advanced
    users feel free to override or overwrite any variables/methods in your subclass.

    Take a look at the pytorch-lightning documentation on how to properly create a
    `LightningModule`:

    https://pytorch-lightning.readthedocs.io/en/stable/common/lightning_module.html

    :::caution

    Ensure you set `self.metrics` in the `__init__` method of your subclass to ensure
    they pertain appropriately to your model. If not, Bitfount will attempt to set
    these appropriately for you but there is no guarantee it will get it right.

    :::

    Args:
        batch_size: The batch size to use for training. Defaults to 32.
        epochs: The number of epochs to train for.
        steps: The number of steps to train for.
        **kwargs: Any additional arguments to pass to parent constructors.

    Attributes:
        batch_size: The batch size to use for training.
        epochs: The number of epochs to train for.
        steps: The number of steps to train for.
        preds: The predictions from the most recent test run.
        target: The targets from the most recent test run.
        val_stats: Metrics from the validation set during training.

    Raises:
        ValueError: If both `epochs` and `steps` are specified.
    """

    train_dl: _BasePyTorchBitfountDataLoader
    # Test attributes
    _test_preds: Optional[List[np.ndarray]]
    _test_targets: Optional[List[np.ndarray]]

    def __init__(
        self,
        batch_size: int = 32,
        epochs: Optional[int] = None,
        steps: Optional[int] = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)

        if (steps is None and epochs is None) or (
            isinstance(steps, int) and isinstance(epochs, int)
        ):
            raise ValueError("You must specify one (and only one) of steps or epochs.")

        # Set hyperparameters
        self.batch_size = batch_size
        self.epochs: Optional[int] = epochs
        self.steps: Optional[int] = steps

        # Set training attributes
        # Override self._model with your model
        self._model: Optional[nn.Module] = None
        self._pl_trainer: pl.Trainer = self.trainer_init()  # cannot be "self.trainer"
        self.preds: List[float] = []
        self.target: List[float] = []
        self.val_stats: List[Dict[str, float]] = []
        self._trained_on_previous_batch: bool = False
        self._total_num_batches_trained: int = 0

    @staticmethod
    def _get_import_statements() -> List[str]:
        """Returns a list of import statements likely to be required for the model.

        Returns:
            A list of import statements.
        """
        return [
            "import os",
            "import torch",
            "from torch import nn as nn",
            "from torch.nn import functional as F",
            "from bitfount import *",
            "import bitfount",
        ]

    @abstractmethod
    def forward(self, x: Any) -> Any:  # type:ignore[override] # Reason: see below
        """Forward method of the model - just like a regular `torch.nn.Module` class.

        :::tip

        This will depend on your model but could be as simple as:

        ```python
        return self._model(x)
        ```

        :::

        Args:
            x: Input to the model.

        Returns:
            Output of the model.
        """
        # pl.LightningModule incompatible with our forward function.
        raise NotImplementedError

    @abstractmethod
    def configure_optimizers(
        self,
    ) -> Union[_OptimizerType, Tuple[List[_OptimizerType], List[_LRScheduler]]]:
        """Configures the optimizer(s) and scheduler(s) for backpropagation.

        Returns:
            Either the optimizer of your choice or a tuple of optimizers and learning
            rate schedulers.
        """
        raise NotImplementedError

    @abstractmethod
    def create_model(self) -> nn.Module:
        """Creates and returns the underlying pytorch model.

        Returns:
            Underlying pytorch model. This is set to `self._model`.
        """
        raise NotImplementedError

    @abstractmethod
    def training_step(self, batch: Any, batch_idx: int) -> _STEP_OUTPUT:  # type: ignore[override] # Reason: see below  # noqa: B950
        """Training step.

        :::caution

        If iterations have been specified in terms of steps, the default behaviour of
        pytorch lightning is to train on the first _n_ steps of the dataloader every
        time `fit` is called. This default behaviour is not desirable and has been dealt
        with in the built-in Bitfount models but, until this bug gets fixed by the
        pytorch lightning team, this needs to be implemented by the user for custom
        models.

        :::

        :::tip

        Take a look at the `skip_training_batch` method for one way on how to deal with
        this. It can be used as follows:

        ```python
        if self.skip_training_batch(batch_idx):
            return None
        ```

        :::

        Args:
            batch: The batch to be trained on.
            batch_idx: The index of the batch to be trained on from the train
                dataloader.

        Returns:
            The loss from this batch as a `torch.Tensor`. Or a dictionary which includes
            the key `loss` and the loss as a `torch.Tensor`.
        """
        # Override the pl.lightning method, as it requires *args
        # and **kwargs as arguments whereas we will always have
        # the batch and the batch_idx as args
        raise NotImplementedError

    @abstractmethod
    def validation_step(self, batch: Any, batch_idx: int) -> _StrAnyDict:  # type: ignore[override] # Reason: see below # noqa: B950
        """Validation step.

        Args:
            batch: The batch to be evaluated.
            batch_idx: The index of the batch to be evaluated from the validation
                dataloader.

        Returns:
            A dictionary of strings and values that should be averaged at the end of
            every epoch and logged e.g. `{"validation_loss": loss}`. These will be
            passed to the `validation_epoch_end` method.
        """
        # Override the pl.lightning method, as it requires *args
        # and **kwargs as arguments whereas we will always have
        # the batch and the batch_idx as args
        raise NotImplementedError

    @abstractmethod
    def test_step(self, batch: Any, batch_idx: int) -> _STEP_OUTPUT:  # type: ignore[override] # Reason: see below  # noqa: B950
        """Operates on a single batch of data from the test set.

        Args:
            batch: The batch to be evaluated.
            batch_idx: The index of the batch to be evaluated from the test
                dataloader.

        Returns:
            A dictionary of predictions and targets, with the dictionary
            keys being "predictions" and "targets" for each of them, respectively.
            These will be passed to the `test_epoch_end` method.
        """
        # Override the pl.lightning method, as it requires
        # *args and **kwargs as arguments whereas we will
        # always have the batch and the batch_idx as args
        raise NotImplementedError

    def tensor_precision(self) -> T_DTYPE:
        """Returns tensor dtype used by Pytorch Lightning Trainer.

        :::note

        Currently only 32-bit training is supported.

        :::

        Returns:
            Pytorch tensor dtype.
        """
        # TODO: [BIT-727] support non-32 bit training
        return cast(T_DTYPE, _TORCH_DTYPES[cast(int, self._pl_trainer.precision)])

    def initialise_model(
        self, data: Optional[BaseSource] = None, context: Optional[ModelContext] = None
    ) -> None:
        """Any initialisation of models/dataloaders to be done here.

        Initialises the dataloaders and sets `self._model` to be the output from
        `self.create_model`. Any initialisation ahead of federated or local training,
        serialization or deserialization should be done here.

        Args:
            data: The datasource for model training. Defaults to None.
            context: Indicates if the model is running as a modeller or worker.
                If None, there is no difference between modeller and worker.
        """
        self._context = context
        self._initialised = True
        if self._context == ModelContext.MODELLER:
            # In a distributed setting, the Modeller needs to first initialise its
            # own model before it can be used. The pod identifier needs to be set
            # before the model is initialised so the the relevant details can be
            # retrieved from the schema. For this we just use the first pod
            # identifier specified in the datastructure as it is assumed the the
            # schemas for all the Pods are the same.
            pod_identifiers = self.datastructure.get_pod_identifiers()
            if pod_identifiers:
                self.set_pod_identifier(pod_identifiers[0])
        if data is not None:
            if self.datastructure.query:
                table_schema = self.datastructure._override_schema(
                    pod_identifier=self.pod_identifier
                )
                self.databunch = BitfountDataBunch(
                    data_structure=self.datastructure,
                    schema=table_schema,
                    datasource=data,
                )
            elif self.datastructure.table:
                if context:
                    table_schema = self.schema.get_table_schema(
                        self.datastructure.get_table_name(self.pod_identifier)
                    )
                    self.databunch = BitfountDataBunch(
                        data_structure=self.datastructure,
                        schema=table_schema,
                        datasource=data,
                    )
                else:
                    # For local training, we can add the datasource to the schema.
                    data.load_data(table_name=self.datastructure.table)
                    self._add_datasource_to_schema(data)
                    table_schema = self.schema.get_table_schema(
                        self.datastructure.get_table_name(self.pod_identifier)
                    )

            if self._context != ModelContext.MODELLER:
                self._set_dataloaders(self.batch_size)
        else:
            if self.datastructure.query:
                table_schema = self.datastructure._override_schema(
                    pod_identifier=self.pod_identifier
                )
            else:
                table_schema = self.schema.get_table_schema(
                    self.datastructure.get_table_name(self.pod_identifier)
                )
        self.datastructure.set_training_input_size(table_schema)

        if hasattr(self, "_objective") and self._objective == "classification":
            # The casts here are to assuage mypy because it (incorrectly) asserts
            # that a subclass of both ClassifierMixIn and BitfountModel cannot exist.
            # We utilise a subclass of both in the tests to assure ourselves.
            if isinstance(cast(ClassifierMixIn, self), ClassifierMixIn):
                cast(ClassifierMixIn, self).set_number_of_classes(table_schema)
            else:
                raise TypeError(
                    "Training objective is classification but this model does not "
                    "inherit from ClassifierMixIn"
                )
        self._model = self.create_model()

    def trainer_init(self) -> pl.Trainer:
        """Initialises the Lightning Trainer for this model.

        Documentation for pytorch-lightning trainer can be found here:
        https://pytorch-lightning.readthedocs.io/en/stable/common/trainer.html

        :::tip

        Override this method to choose your own `Trainer` arguments.

        :::

        Returns:
            The pytorch lightning trainer.
        """
        callbacks: List[Callback] = [TQDMProgressBar(refresh_rate=1)]

        trainer = pl.Trainer(
            max_epochs=self.epochs or -1,
            max_steps=self.steps or -1,
            deterministic=True,
            auto_lr_find=True,
            gpus=_autodetect_gpu(),  # will be 1 (first GPU) or 0 (CPU)
            callbacks=callbacks,
        )
        return trainer

    def train_dataloader(self) -> _BasePyTorchBitfountDataLoader:  # type: ignore[override] # Reason: see below # noqa: B950
        """Returns training dataloader."""
        # We override the dataloader return annotation as the LightningModule
        # expects a pytorch DataLoader, and we return out PyTorchBitfountDataLoader
        return self.train_dl

    def val_dataloader(self) -> _BasePyTorchBitfountDataLoader:  # type: ignore[override] # Reason: see below # noqa: B950
        """Returns validation dataloader."""
        # We override the dataloader return annotation as the LightningModule
        # expects a pytorch DataLoader, and we return out PyTorchBitfountDataLoader
        return cast(_BasePyTorchBitfountDataLoader, self.validation_dl)

    def test_dataloader(self) -> _BasePyTorchBitfountDataLoader:  # type: ignore[override] # Reason: see below # noqa: B950
        """Returns test dataloader."""
        # We override the dataloader return annotation as the LightningModule
        # expects a pytorch DataLoader, and we return out PyTorchBitfountDataLoader
        return cast(_BasePyTorchBitfountDataLoader, self.test_dl)

    def serialize(self, filename: Union[str, os.PathLike]) -> None:
        """Serialize model to file with provided `filename`.

        Args:
            filename: Path to file to save serialized model.
        """
        if not self._initialised:
            logger.info("Model not yet initialised. Auto-initialising model.")
            self.initialise_model()
        # Model has been initialised, assuring mypy of this
        assert self._model is not None  # nosec[assert_used]
        torch.save(self._model.state_dict(), filename)

    def deserialize(self, filename: Union[str, os.PathLike]) -> None:
        """Deserialize model.

        :::danger

        This should not be used on a model file that has been received across a
        trust boundary due to underlying use of `pickle` by `torch`.

        :::

        Args:
            filename: Path to file containing serialized model.
        """
        if not self._initialised:
            logger.info("Model not yet initialised. Auto-initialising model.")
            self.initialise_model()
        # Model has been initialised, assuring mypy of this
        assert self._model is not None  # nosec[assert_used]
        self._model.load_state_dict(torch.load(filename))

    def skip_training_batch(self, batch_idx: int) -> bool:
        """Checks if the current batch from the training set should be skipped.

        This is a workaround for the fact that PyTorch Lightning starts the Dataloader
        iteration from the beginning every time `fit` is called. This means that if we
        are training in steps, we are always training on the same batches. So this
        method needs to be called at the beginning of every `training_step` to skip
        to the right batch index.

        Args:
            batch_idx: the index of the batch from `training_step`.

        Returns:
            True if the batch should be skipped, otherwise False.
        """
        # TODO: [BIT-1237] remove this code block and find a better way to do this that
        # doesn't involve loading every batch into memory until we get to the right one
        if self.steps:
            # If we have trained on the previous batch, we can avoid the checks because
            # it means we have already reached the target start batch.
            if not self._trained_on_previous_batch:
                if (self.steps != self._pl_trainer.max_steps) and (
                    batch_idx < (self._total_num_batches_trained % len(self.train_dl))
                ):
                    return True
                else:
                    self._trained_on_previous_batch = True

            # `_total_num_batches_trained` hasn't been incremented yet so we need to add
            # 1 here to get the correct batch number.
            if self._total_num_batches_trained + 1 == self._pl_trainer.max_steps:
                self._trained_on_previous_batch = False

        if not self._pl_trainer.sanity_checking:
            self._total_num_batches_trained += 1

        return False

    def validation_epoch_end(  # type: ignore[override] # Reason: see below
        self, outputs: List[_StrAnyDict]
    ) -> None:
        """Called at the end of the validation epoch with all validation step outputs.

        Ensures that the average metrics from a validation epoch is stored. Logs results
        and also appends to `self.val_stats`.

        Args:
            outputs: List of outputs from each validation step.
        """
        # Override the pl.lightning method, as its outputs can be
        # List[Union[Tensor, _StrAnyDict]],
        # whereas we force outputs to be a Dict
        avgs = self._compute_metric_averages(outputs)
        self.val_stats.append(avgs)

        # Also log out these averaged metrics
        for k, v in avgs.items():
            self.log(f"avg_{k}", v)

    def test_epoch_end(self, outputs: List[Dict[str, torch.Tensor]]) -> None:  # type: ignore[override] # Reason: see below # noqa: B950
        """Aggregates the predictions and targets from the test set.

        Args:
            outputs: List of outputs from each test step.
        """
        # Override the pl.lightning method, as it requires a different type for outputs.
        merged_outputs: Dict[str, List[torch.Tensor]] = _merge_list_of_dicts(outputs)
        self._test_preds = [i.cpu().numpy() for i in merged_outputs["predictions"]]
        self._test_targets = [i.cpu().numpy() for i in merged_outputs["targets"]]

    @staticmethod
    def _compute_metric_averages(outputs: List[_StrAnyDict]) -> Dict[str, float]:
        """Compute the average metrics from a list of outputs."""
        # Stack up shared dict keys into lists of entries
        stacked_dict = defaultdict(list)
        for output in outputs:
            for k, v in output.items():
                stacked_dict[k].append(v)

        # Calculate the average value of each key and convert to float
        avgs = {}
        for k, v_list in stacked_dict.items():
            avgs[k] = float(torch.stack(v_list).mean().item())
        return avgs

    def _evaluate_local(
        self, test_dl: Optional[BitfountDataLoader] = None, **kwargs: Any
    ) -> Tuple[np.ndarray, np.ndarray]:
        """This method runs inference on the test dataloader.

        This is done by calling `self.test_step` under the hood. Customise this method
        as you please but it must return a list of predictions and a list of targets.

        Args:
            test_dl: Optional dataloader to run inference on which takes precedence over
                the dataloader returned by `self.test_dataloader`.

        Returns:
            A tuple of predictions and targets as numpy arrays.
        """
        # Reset test attributes to None
        self._reset_test_attrs()
        if test_dl is None:
            if isinstance(self.test_dl, _BasePyTorchBitfountDataLoader):
                test_dl = self.test_dl
            else:
                raise ValueError("No test data to evaluate the model on.")
        self._pl_trainer.test(model=self, dataloaders=cast(PyTorchDataLoader, test_dl))
        return np.asarray(self._test_preds), np.asarray(self._test_targets)

    def _reset_test_attrs(self) -> None:
        """Resets test attributes to None."""
        self._test_preds = None
        self._test_targets = None

    def predict(self, data: BaseSource, **kwargs: Any) -> np.ndarray:
        """This method runs inference on the test data, returns predictions.

        This is done by calling `test_step` under the hood. Customise this method as you
        please but it must return a list of predictions and a list of targets. Note that
        as this is the prediction function, only the predictions are returned.

        Returns:
            A numpy array containing the prediction values.

        Raises:
            ValueError: If no data is provided to test with.
        """
        data.load_data()
        self._add_datasource_to_schema(datasource=data)
        test_dl = self.databunch.get_test_dataloader(self.batch_size)
        if isinstance(test_dl, BitfountDataLoader):
            logger.info(
                f"Using test portion of dataset for inference - this has"
                f"{len(test_dl)} records."
            )
        else:
            raise ValueError("No data provided to run inference for the model on.")

        self._pl_trainer.test(model=self, dataloaders=cast(PyTorchDataLoader, test_dl))
        return cast(np.ndarray, self._test_preds)

    def _fit_local(
        self,
        data: BaseSource,
        metrics: Optional[Union[str, List[str], MutableMapping[str, Metric]]] = None,
        **kwargs: Any,
    ) -> Dict[str, str]:
        """Trains the model on local data.

        Returns:
            Validation metrics for the final epoch.
        """
        if not self._initialised:
            logger.info("Model not yet initialised. Auto-initialising model.")
            self.initialise_model(data)

        self._pl_trainer.fit(self)

        # Return the validation stats to be sent back
        return {k: ("%.4f" % v) for k, v in self.val_stats[-1].items()}



"""Classes concerning datasets."""
from __future__ import annotations

from abc import ABC, abstractmethod
from functools import cached_property
import logging
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Iterator,
    List,
    Literal,
    Mapping,
    Optional,
    OrderedDict,
    Sequence,
    Tuple,
    Union,
    cast,
)

import numpy as np
import pandas as pd
from skimage import color, io
from sqlalchemy import text

from bitfount.data.datasources.base_source import BaseSource
from bitfount.data.datasources.database_source import DatabaseSource
from bitfount.data.datasplitters import DatasetSplitter, PercentageSplitter
from bitfount.data.exceptions import DataNotLoadedError
from bitfount.data.types import DataSplit
from bitfount.transformations.base_transformation import Transformation
from bitfount.transformations.processor import TransformationProcessor
from bitfount.utils import _array_version

if TYPE_CHECKING:

    from bitfount.data.schema import TableSchema
    from bitfount.data.types import (
        _DataEntry,
        _ImagesData,
        _SemanticTypeValue,
        _SupportData,
        _TabularData,
    )
    from bitfount.transformations.batch_operations import BatchTimeOperation

logger = logging.getLogger(__name__)


class _BaseBitfountDataset(ABC):
    """Base class for representing a dataset."""

    x_columns: List[str]
    x_var: Tuple[Any, Any, np.ndarray]
    y_columns: List[str]
    y_var: np.ndarray

    embedded_col_names: List[str]
    image_columns: List[str]
    processors: Dict[int, TransformationProcessor]
    image: np.ndarray
    tabular: np.ndarray
    support_cols: np.ndarray

    def __init__(
        self,
        datasource: BaseSource,
        data_split: DataSplit,
        schema: TableSchema,
        selected_cols: List[str],
        selected_cols_semantic_types: Mapping[_SemanticTypeValue, List[str]],
        data_splitter: Optional[DatasetSplitter] = None,
        target: Optional[Union[str, List[str]]] = None,
        batch_transforms: Optional[List[BatchTimeOperation]] = None,
        weights_col: Optional[str] = None,
        multihead_col: Optional[str] = None,
        ignore_classes_col: Optional[str] = None,
    ) -> None:
        super().__init__()
        self.datasource = datasource
        self.schema = schema
        self.selected_cols = selected_cols
        self.selected_cols_semantic_types = selected_cols_semantic_types
        self.data_splitter = data_splitter
        self.target = target
        self.batch_transforms = batch_transforms
        self.data_split = data_split
        self.weights_col = weights_col
        self.multihead_col = multihead_col
        self.ignore_classes_col = ignore_classes_col

        self._set_column_name_attributes()
        self._set_batch_transformation_processors()

    @abstractmethod
    def __len__(self) -> int:
        raise NotImplementedError

    def _apply_schema(self, data: pd.DataFrame) -> None:
        """Applies `self.schema` to `data` and sets the result to `self.data`.

        `selected_cols` needs to be passed to the `apply` method here to ensure
        that we don't end up removing the extra columns in our dataframe that are
        used during training (e.g. loss_weights_col, etc.) but aren't part of the
        schema. Applying the schema adds extra columns to the dataframe if they
        are missing. Therefore we need to subset the data columns here to ensure
        we are only using the columns specified for this task
        """
        diff = list(sorted(set(self.selected_cols) - set(data.columns)))
        if diff:
            logger.warning(
                f"Selected columns `{','.join(diff)}` "
                f"were not found in the data, continuing without them."
            )
            self.selected_cols = [i for i in self.selected_cols if i not in diff]

        self.data = self.schema.apply(data, keep_cols=self.selected_cols)[
            self.selected_cols
        ].reset_index(drop=True)

    def _set_column_name_attributes(self) -> None:
        """Sets the attributes concerning column names.

        Namely, `self.x_columns`, `self.y_columns`, `self.embedded_col_names`,
        and `self.image_columns`.
        """
        self.image_columns = self.selected_cols_semantic_types.get("image", [])
        self.embedded_col_names = self.selected_cols_semantic_types.get(
            "categorical", []
        )
        self.x_columns = (
            self.embedded_col_names
            + self.selected_cols_semantic_types.get("continuous", [])
            + self.selected_cols_semantic_types.get("image", [])
        )
        if self.target is not None:
            self.y_columns = _array_version(self.target)
            self.embedded_col_names = [
                i for i in self.embedded_col_names if i not in self.y_columns
            ]
            self.x_columns = [i for i in self.x_columns if i not in self.y_columns]

    def _set_batch_transformation_processors(self) -> None:
        """Sets `self.processors` for batch transformations."""
        if self.batch_transforms is not None:
            # We create a dictionary mapping each image feature to the corresponding
            # list of transformations. This dictionary must be an OrderedDict so that
            # the order of the features is preserved and indexable. Currently, we only
            # support image transformations at batch time.
            feature_transforms: OrderedDict[
                str, List[BatchTimeOperation]
            ] = OrderedDict(
                {i: [] for i in self.selected_cols_semantic_types.get("image", [])}
            )

            for tfm in self.batch_transforms:
                if tfm.arg in feature_transforms:
                    feature_transforms[tfm.arg].append(tfm)

            # Each feature that will be transformed needs to have its own transformation
            # processor. These processors need to correspond to the index of the feature
            # to be transformed because at batch time, the feature name is unavailable -
            # we only have the feature index. Finally, we only leave transformations if
            # the 'step' corresponds to the 'step' of the Dataset. This is to optimise
            # for efficiency only since the processor will ignore transformations that
            # are not relevant to the current step at batch time anyway.
            self.processors: Dict[int, TransformationProcessor] = {
                list(feature_transforms).index(col): TransformationProcessor(
                    [
                        cast(Transformation, i)
                        for i in tfms
                        if i.step == self.data_split
                    ],
                )
                for col, tfms in feature_transforms.items()
            }

    def _transform_image(self, img: np.ndarray, idx: int) -> np.ndarray:
        """Performs image transformations if they have been specified.

        Args:
            img: The image to be transformed.
            idx: The index of the image.

        Returns:
            The transformed image.

        """
        if not self.batch_transforms:
            return img

        return self.processors[idx].batch_transform(img, step=self.data_split)

    def _load_images(
        self,
        idx: Union[int, Sequence[int]],
        what_to_load: Literal["target", "image"] = "image",
    ) -> Union[np.ndarray, Tuple[np.ndarray, ...]]:
        """Loads images and performs transformations if specified.

        This involves first converting grayscale images to RGB if necessary.

        Args:
            idx: The index to be loaded.
            what_to_load: Str variable specifying whether to load 'image' or 'target'.

        Returns:
            Loaded and transformed image.

        """
        if what_to_load == "image":
            img_features = self.image[idx]
        elif what_to_load == "target":
            img_features = np.array([self.y_var[idx]])
        imgs: Tuple[np.ndarray, ...] = tuple(
            io.imread(image, plugin="pil") for image in img_features
        )
        imgs = tuple(
            color.gray2rgb(image_array) if len(image_array.shape) < 3 else image_array
            for image_array in imgs
        )
        imgs = tuple(
            self._transform_image(image_array, i) for i, image_array in enumerate(imgs)
        )

        if len(img_features) == 1:
            return imgs[0]

        return imgs

    def _set_support_column_values(self, data: pd.DataFrame) -> None:
        """Sets `self.support_cols` - auxiliary columns for loss manipulation."""
        if self.weights_col:
            weights = data.loc[:, [self.weights_col]].values.astype(np.float32)
            self.x_columns.append(self.weights_col)
        else:
            weights = np.ones(len(data), dtype=np.float32)
        weights = weights.reshape(len(weights), 1)

        if self.ignore_classes_col:
            ignore_classes = data.loc[:, [self.ignore_classes_col]].values.astype(
                np.int64
            )
        else:
            ignore_classes = -np.ones(len(data), dtype=np.int64)
        ignore_classes = ignore_classes.reshape(len(ignore_classes), 1)

        if self.multihead_col:
            category = data.loc[:, [self.multihead_col]].values
            category = category.reshape(len(category), 1)
            self.support_cols = cast(
                np.ndarray, np.concatenate((weights, ignore_classes, category), axis=1)
            )
        else:
            self.support_cols = cast(
                np.ndarray, np.concatenate((weights, ignore_classes), axis=1)
            )

    def _set_image_values(self, data: pd.DataFrame) -> None:
        """Sets `self.image`."""
        if self.image_columns != []:
            for (i, col) in enumerate(self.image_columns):
                x_img = np.expand_dims(data.loc[:, col].values, axis=1)
                if i == 0:
                    self.image = x_img

                else:
                    self.image = np.concatenate((self.image, x_img), axis=1)
        else:
            self.image = np.array([])

    def _set_tabular_values(self, data: pd.DataFrame) -> None:
        """Sets `self.tabular`."""
        x1_var = data.loc[:, self.embedded_col_names].values.astype(np.int64)
        x2_var = data.loc[
            :, self.selected_cols_semantic_types.get("continuous", [])
        ].values.astype(np.float32)
        self.tabular = np.concatenate((x1_var, x2_var), axis=1)

    def _set_target_values(
        self, target: Optional[Union[pd.DataFrame, pd.Series]]
    ) -> None:
        """Sets `self.y_var`."""
        if target is not None:
            self.y_var = cast(np.ndarray, target.values)
        else:
            self.y_var = np.array([])

    def _get_xy(
        self, data: pd.DataFrame
    ) -> Tuple[pd.DataFrame, Optional[Union[pd.DataFrame, pd.Series]]]:
        """Returns the x and y variables.

        By default, there is no target unless `self.target` has been specified.
        """
        X, Y = data, None

        if self.target is not None:
            # ignore error if target is already not part of the X data
            X = X.drop(columns=self.target, errors="ignore").reset_index(drop=True)
            Y = data[self.target].reset_index(drop=True)
        return X, Y

    def _getitem(self, idx: Union[int, Sequence[int]]) -> _DataEntry:
        """Returns the item referenced by index `idx` in the data."""
        image: _ImagesData
        tab: _TabularData
        sup: _SupportData

        target: Union[np.ndarray, Tuple[np.ndarray, ...]]
        if len(self.y_var) == 0:
            # Set the target, if the dataset has no supervision,
            # choose set the default value to be 0.
            target = np.array(0)
        elif (
            "image" in self.schema.features
            and self.target in self.schema.features["image"]
        ):
            # Check if the target is an image and load it.
            target = self._load_images(idx, what_to_load="target")
        else:
            target = self.y_var[idx]

        # If the Dataset contains both tabular and image data
        if self.image.size and self.tabular.size:
            tab = self.tabular[idx]
            sup = self.support_cols[idx]
            image = self._load_images(idx)
            return (tab, image, sup), target

        # If the Dataset contains only tabular data
        elif self.tabular.size:
            tab = self.tabular[idx]
            sup = self.support_cols[idx]
            return (tab, sup), target

        # If the Dataset contains only image data
        else:
            sup = self.support_cols[idx]
            image = self._load_images(idx)
            return (image, sup), target

    def _reformat_data(self, data: pd.DataFrame) -> None:
        """Reformats the data to be compatible with the Dataset class."""
        self._apply_schema(data)
        X, Y = self._get_xy(self.data)

        self._set_image_values(X)
        self._set_tabular_values(X)
        self._set_support_column_values(X)
        # Package tabular, image and support columns together under the x_var attribute
        self.x_var = (self.tabular, self.image, self.support_cols)
        self._set_target_values(Y)

    def _resolve_data_splitter(
        self,
        data_splitter: Optional[DatasetSplitter] = None,
    ) -> DatasetSplitter:
        """Resolves the data splitter.

        The data splitter is resolved in the following order:
            1. BaseSource data_splitter if specified
            2. Provided data_splitter if specified
            3. PercentageSplitter (default)

        Returns:
            The appropriate data splitter to use.
        """
        if self.datasource.data_splitter:
            if data_splitter:
                logger.warning(
                    "Ignoring provided data splitter as the BaseSource "
                    "already has one."
                )
            data_splitter = self.datasource.data_splitter
        elif not data_splitter:
            logger.warning(
                "No data splitter provided. Using default PercentageSplitter."
            )
            data_splitter = PercentageSplitter()

        return data_splitter

    def _split_data(self, data_splitter: Optional[DatasetSplitter] = None) -> None:
        """Split the data into training, validation and test datasets.

        This method is idempotent so it can be called multiple times without
        re-splitting the data.

        Args:
            data_splitter: An optional data splitter object.
        """
        if not self.datasource._data_is_split:
            data_splitter = self._resolve_data_splitter(data_splitter)
            assert self.datasource.data is not None  # nosec[assert_used]

            (
                self.datasource._train_idxs,
                self.datasource._validation_idxs,
                self.datasource._test_idxs,
            ) = data_splitter.create_dataset_splits(self.datasource.data)

            self.datasource._data_is_split = True
        else:
            logger.debug("Data is already split, keeping the current split.")

    def get_dataset_split_length(
        self, split: DataSplit, data_splitter: Optional[DatasetSplitter] = None
    ) -> int:
        """Returns the length of the specified dataset split.

        Args:
            split: The split to get the length of.
            data_splitter: The splitter object used to split the data if the BaseSource
                does not have one.

        Returns:
            The length of the specified dataset split.

        Raises:
            DataNotLoadedError: If unable to get the length of the dataset split.
        """
        data_splitter = self._resolve_data_splitter(data_splitter)

        # If PercentageSplitter is used regardless of the data loader.
        if isinstance(data_splitter, PercentageSplitter):
            len_loader = len(self.datasource)
            if split == DataSplit.TRAIN:
                return int(len_loader * data_splitter.train_percentage / 100)
            elif split == DataSplit.VALIDATION:
                return int(len_loader * data_splitter.validation_percentage / 100)
            elif split == DataSplit.TEST:
                return int(len_loader * data_splitter.test_percentage / 100)

        # If the data is held in memory with any other splitter.
        elif self.datasource._data_is_loaded:
            return len(self.get_dataset_split(split=split, data_splitter=data_splitter))

        # If the data is held in a database with any other splitter. This is because
        # it requires the database to be queried in order to determine the length of
        # a split e.g. `SplitterDefinedInData`
        elif isinstance(self.datasource, DatabaseSource):
            query = self._resolve_data_splitter(data_splitter).get_split_query(
                self.datasource, split
            )
            with self.datasource.con.connect() as con:
                # Ignoring the security warning because the sql query is trusted and
                # will be executed regardless.
                result = con.execute(text(f"SELECT COUNT(*) FROM ({query}) q"))  # nosec
                return cast(int, result.scalar_one())

        # `load_data` should be called to avoid this error being raised
        raise DataNotLoadedError("Unable to get length of dataset split")

    def get_dataset_split(
        self,
        split: DataSplit,
        data_splitter: Optional[DatasetSplitter] = None,
    ) -> pd.DataFrame:
        """Returns the relevant portion of `self.data`.

        Args:
            split: The portion of data to return.
            data_splitter: The splitter object used to split the data.

        Returns:
            A dataframe-type object containing the data points specified by the data
            splitter.

        Raises:
            DataNotLoadedError: If data has not been loaded.
        """
        if not self.datasource._data_is_loaded:
            raise DataNotLoadedError(
                "Please load data before accessing a split. "
                "If the data is remote, it must be iterated."
            )

        self._split_data(data_splitter)  # idempotent
        indices: np.ndarray = getattr(self.datasource, f"_{split.value}_idxs")
        df: pd.DataFrame = self.datasource.data.loc[indices.tolist()]
        return df.reset_index(drop=True)


class _BitfountDataset(_BaseBitfountDataset):
    """A dataset for supervised tasks.

    When indexed, returns numpy arrays corresponding to
    categorical features, continuous features, weights and target value (and
    optionally category)
    """

    def __init__(
        self,
        datasource: BaseSource,
        data_split: DataSplit,
        schema: TableSchema,
        selected_cols: List[str],
        selected_cols_semantic_types: Mapping[_SemanticTypeValue, List[str]],
        data_splitter: Optional[DatasetSplitter] = None,
        target: Optional[Union[str, List[str]]] = None,
        batch_transforms: Optional[List[BatchTimeOperation]] = None,
        weights_col: Optional[str] = None,
        multihead_col: Optional[str] = None,
        ignore_classes_col: Optional[str] = None,
    ) -> None:
        super().__init__(
            datasource=datasource,
            data_split=data_split,
            schema=schema,
            selected_cols=selected_cols,
            selected_cols_semantic_types=selected_cols_semantic_types,
            data_splitter=data_splitter,
            target=target,
            batch_transforms=batch_transforms,
            weights_col=weights_col,
            multihead_col=multihead_col,
            ignore_classes_col=ignore_classes_col,
        )

        data = self.get_dataset_split(
            split=self.data_split, data_splitter=self.data_splitter
        )
        self._reformat_data(data)

    def __len__(self) -> int:
        return len(self.x_var[0])


class _IterableBitfountDataset(_BaseBitfountDataset):
    """Iterable Dataset.

    Currently, this is only used for Database connections.
    """

    def __iter__(self) -> Iterator[_DataEntry]:
        """Iterates over the dataset."""
        for data_partition in self.yield_dataset_split(
            split=self.data_split, data_splitter=self.data_splitter
        ):
            self._reformat_data(data_partition)

            for idx in range(len(self.data)):
                yield self._getitem(idx)

    @cached_property
    def _len(self) -> int:
        """Returns the length of the dataset."""
        return self.get_dataset_split_length(
            split=self.data_split, data_splitter=self.data_splitter
        )

    def __len__(self) -> int:
        return self._len

    def yield_dataset_split(
        self,
        split: DataSplit,
        data_splitter: Optional[DatasetSplitter] = None,
    ) -> Iterator[pd.DataFrame]:
        """Returns an iterator over the relevant data split.

        Args:
            split: The portion of data to yield from.
            data_splitter: The splitter object used to split the data.

        Returns:
            A iterator of pandas dataframes containing the relevant data split.

        Raises:
            ValueError: If no query or table name has been supplied for multi-table
                data.
            ValueError: If attempting to iterate over in-memory data.
        """
        if isinstance(self.datasource, DatabaseSource):
            if not self.datasource.query:
                raise ValueError("No query or table name specified.")

            query = self._resolve_data_splitter(data_splitter).get_split_query(
                self.datasource, split
            )
            with self.datasource.con.connect() as con:
                result = con.execute(text(query))
                for partition in result.partitions(
                    self.datasource.database_partition_size
                ):
                    yield pd.DataFrame(partition, columns=list(result.keys()))
        else:
            raise ValueError(
                "Iterating over in-memory data is currently not supported."
            )



"""Contains functionality for interacting with the task monitoring service.

A monitor module is available globally when running within a task's context (created
by task_monitor_context()); outside of this context calls to the monitor module
will fail as no task is in progress and hence no task ID is available.

Attempting to instantiate a second monitor module (i.e. running a second task context)
will raise an exception.

Interacting with the monitor module is achieved by utilising the appropriate functions
to send a desired update. These will only work within a task context.

Decorator versions of stateless functions are available in monitoring.decorators;
these can be applied to functions/methods and will make the appropriate monitor
service update when those functions or methods are called.
"""

from bitfount.federated.monitoring.exceptions import (
    ExistingMonitorModuleError,
    MonitorModuleError,
    NoMonitorModuleError,
)
from bitfount.federated.monitoring.monitor import (
    task_config_update,
    task_monitor_context,
    task_status_update,
)
from bitfount.federated.monitoring.types import (
    AdditionalMonitorMessageTypes,
    HasTaskID,
    MonitorRecordPrivacy,
    ProgressCounterDict,
)

__all__ = [
    "ExistingMonitorModuleError",
    "MonitorModuleError",
    "NoMonitorModuleError",
    "task_config_update",
    "task_monitor_context",
    "task_status_update",
    "AdditionalMonitorMessageTypes",
    "HasTaskID",
    "MonitorRecordPrivacy",
    "ProgressCounterDict",
]

__pdoc__ = {}
for _obj in __all__:
    __pdoc__[_obj] = False



"""PyTorch data utility functions to be found here."""
from typing import List, Sequence, Union, cast

import numpy as np
import torch

from bitfount.data.types import _DataBatch, _SingleOrMulti


def _index_tensor_handler(
    idx: Union[int, Sequence[int], torch.Tensor]
) -> Union[int, Sequence[int]]:
    """Converts pytorch tensors to integers or lists of integers for indexing."""
    if torch.is_tensor(idx):
        idx = cast(torch.Tensor, idx)
        list_idx: list = idx.tolist()
        return list_idx
    else:
        idx = cast(Union[int, Sequence[int]], idx)
        return idx


def _convert_batch_to_tensor(
    batch: _DataBatch,
) -> List[_SingleOrMulti[torch.Tensor]]:
    """Converts a batch of data containing numpy arrays to torch tensors.

    Data must be explicitly converted to torch tensors since the PyTorch DataLoader
    which does this automatically is not being used.
    """
    x = [torch.tensor([b[0][i] for b in batch]) for i in range(len(batch[0][0]))]
    y = torch.from_numpy(np.array([b[1] for b in batch]))

    return [x, y]



"""Tests for the backends/pytorch/loss.py."""

import numpy as np
import pytest
import torch

from bitfount.backends.pytorch.loss import SoftDiceLoss, soft_dice_loss
from tests.utils.helper import backend_test, unit_test


@backend_test
@unit_test
class TestSoftDiceLoss:
    """Tests the SoftDiceLoss."""

    def test_softdiceloss(self) -> None:
        """Test the SoftDiceLoss."""
        loss_fn = SoftDiceLoss()
        inp = torch.zeros((1, 3, 1, 1), requires_grad=True) + 0.1
        target = torch.zeros((1, 1, 1), dtype=torch.long) + 1
        loss = loss_fn(inp, target)
        loss.backward()
        assert isinstance(loss, torch.Tensor)
        # The value for the second tensor taken from the test for the
        # original implementation.
        assert torch.isclose(loss, torch.tensor(-0.79653674364089))

    def test_softdiceloss_with_weights(self) -> None:
        """Test the SoftDiceLoss with predefined weights."""
        loss_fn = SoftDiceLoss(weight=[0.33, 0.33, 0.33])
        assert loss_fn.weight is not None

    def test_softdiceloss_square_nom_denom(self) -> None:
        """Test the SoftDiceLoss with square_nom and denom."""
        loss_fn = SoftDiceLoss(square_nom=True, square_denom=True)
        inp = torch.zeros((1, 3, 1), requires_grad=True) + 0.1
        target = torch.zeros((1, 1), dtype=torch.long) + 1
        loss = loss_fn(inp, target)
        loss.backward()
        assert isinstance(loss, torch.Tensor)
        assert torch.isclose(loss, torch.tensor(-0.8292231))

    def test_softdiceloss_error_ndim_targets(self) -> None:
        """Test the SoftDiceLoss raises error with targetsndim less than 3."""
        loss_fn = SoftDiceLoss(square_nom=True, square_denom=True)
        inp = torch.zeros((1, 3, 1), requires_grad=True) + 0.1
        target = torch.zeros((1,), dtype=torch.long) + 1
        with pytest.raises(ValueError):
            loss_fn(inp, target)

    def test_softdiceloss_error_ndim_preds(self) -> None:
        """Test the SoftDiceLoss raises error with predictions.ndim less than 3."""
        loss_fn = SoftDiceLoss(square_nom=True, square_denom=True)
        inp = torch.zeros((1, 3), requires_grad=True) + 0.1
        target = torch.zeros((1, 1), dtype=torch.long) + 1
        with pytest.raises(ValueError):
            loss_fn(inp, target)


@backend_test
@unit_test
def test_func_soft_dice_loss() -> None:
    """Tests functional implementation of soft dice loss."""
    inp = np.zeros((1, 3, 1)) + 0.1
    target = np.zeros((1, 1)) + 1
    loss = soft_dice_loss(inp, target)
    assert isinstance(loss, torch.Tensor)
    assert torch.isclose(loss, torch.tensor([-0.7381], dtype=torch.float64))



"""Utility functions for aggregator tests."""
from typing import Mapping, Union

import numpy as np
from numpy.testing import assert_array_equal

from bitfount.backends.pytorch.types import _AdaptorForPyTorchTensor
from bitfount.types import _Weights


def assert_equal_weight_dicts(
    a: Union[_Weights, Mapping[str, np.ndarray]],
    b: Union[_Weights, Mapping[str, np.ndarray]],
) -> None:
    """Checks if two WeightDicts are equal.

    Check keys match and array values match.
    """
    # Check keys
    assert a.keys() == b.keys()
    # Check arrays
    for key, value in a.items():
        if isinstance(value, _AdaptorForPyTorchTensor):
            assert_array_equal(np.asarray(value.torchtensor), np.asarray(b[key]))
        else:
            assert_array_equal(np.asarray(value), np.asarray(b[key]))



"""Fixtures for pytorch data tests."""

import pandas as pd
from pytest import fixture

from bitfount.backends.pytorch.data.datasets import (
    _PyTorchDataset,
    _PyTorchIterableDataset,
)
from bitfount.data.datasources.dataframe_source import DataFrameSource
from bitfount.data.datastructure import DataStructure
from bitfount.data.schema import BitfountSchema
from bitfount.data.types import DataSplit, SemanticType
from tests.utils.helper import TABLE_NAME, create_dataset


@fixture
def dataframe() -> pd.DataFrame:
    """Underlying dataframe for dataset."""
    return create_dataset(image=True)


@fixture
def tabular_dataset(dataframe: pd.DataFrame) -> _PyTorchDataset:
    """Basic PyTorch tabular dataset for tests as fixture."""
    target = "TARGET"
    datasource = DataFrameSource(dataframe)
    datasource.load_data()

    schema = BitfountSchema()
    schema.add_datasource_tables(
        datasource,
        table_name=TABLE_NAME,
        force_stypes={TABLE_NAME: {"categorical": ["Date"]}},
    )
    datasource.data = datasource.data.drop(
        columns=schema.get_feature_names(TABLE_NAME, SemanticType.TEXT)
    )
    datastructure = DataStructure(
        target=target, ignore_cols=["image"], table=TABLE_NAME
    )
    datastructure.set_training_column_split_by_semantic_type(schema.tables[0])
    return _PyTorchDataset(
        datasource=datasource,
        target=target,
        selected_cols_semantic_types=datastructure.selected_cols_w_types,
        selected_cols=datastructure.selected_cols,
        data_split=DataSplit.TRAIN,
        schema=schema.get_table_schema(TABLE_NAME),
    )


@fixture
def iter_tabular_dataset(dataframe: pd.DataFrame) -> _PyTorchIterableDataset:
    """Iterable PyTorch tabular dataset for tests as fixture."""
    target = "TARGET"
    datasource = DataFrameSource(dataframe)
    datasource.load_data()

    schema = BitfountSchema()
    schema.add_datasource_tables(
        datasource,
        table_name=TABLE_NAME,
        force_stypes={TABLE_NAME: {"categorical": ["Date"]}},
    )
    datasource.data = datasource.data.drop(
        columns=schema.get_feature_names(TABLE_NAME, SemanticType.TEXT)
    )
    datastructure = DataStructure(
        target=target, ignore_cols=["image"], table=TABLE_NAME
    )
    datastructure.set_training_column_split_by_semantic_type(schema.tables[0])
    return _PyTorchIterableDataset(
        datasource=datasource,
        target=target,
        selected_cols_semantic_types=datastructure.selected_cols_w_types,
        selected_cols=datastructure.selected_cols,
        data_split=DataSplit.TRAIN,
        schema=schema.get_table_schema(TABLE_NAME),
    )


@fixture
def image_dataset(dataframe: pd.DataFrame) -> _PyTorchDataset:
    """Basic PyTorch image dataset for tests as fixture."""
    target = "TARGET"
    datasource = DataFrameSource(dataframe)
    schema = BitfountSchema()
    schema.add_datasource_tables(
        datasource,
        force_stypes={TABLE_NAME: {"image": ["image"], "categorical": ["Date"]}},
        table_name=TABLE_NAME,
    )
    datasource.load_data()
    datasource.data = datasource.data.drop(
        columns=schema.get_feature_names(TABLE_NAME, SemanticType.TEXT)
    )
    datastructure = DataStructure(
        target=target, selected_cols=["image", "TARGET"], table=TABLE_NAME
    )
    datastructure.set_training_column_split_by_semantic_type(schema.tables[0])
    return _PyTorchDataset(
        datasource=datasource,
        target=target,
        selected_cols_semantic_types=datastructure.selected_cols_w_types,
        selected_cols=datastructure.selected_cols,
        schema=schema.get_table_schema(TABLE_NAME),
        data_split=DataSplit.TRAIN,
    )


@fixture
def image_tab_dataset(dataframe: pd.DataFrame) -> _PyTorchDataset:
    """Basic PyTorchPredictionDataset dataset for tests as fixture."""
    target = "TARGET"
    datasource = DataFrameSource(dataframe)
    schema = BitfountSchema()
    schema.add_datasource_tables(
        datasource,
        force_stypes={TABLE_NAME: {"image": ["image"], "categorical": ["Date"]}},
        table_name=TABLE_NAME,
    )
    datasource.load_data()
    datasource.data = datasource.data.drop(
        columns=schema.get_feature_names(TABLE_NAME, SemanticType.TEXT)
    )
    datastructure = DataStructure(target=target, table=TABLE_NAME)
    datastructure.set_training_column_split_by_semantic_type(schema.tables[0])
    return _PyTorchDataset(
        datasource=datasource,
        target=target,
        selected_cols_semantic_types=datastructure.selected_cols_w_types,
        selected_cols=datastructure.selected_cols,
        schema=schema.get_table_schema(TABLE_NAME),
        data_split=DataSplit.TRAIN,
    )


@fixture
def multiimage_dataframe() -> pd.DataFrame:
    """Underlying dataframe for multi-image dataset."""
    return create_dataset(multiimage=True)


@fixture
def multiimage_dataset(multiimage_dataframe: pd.DataFrame) -> _PyTorchDataset:
    """Basic multi-image dataset for tests as fixture."""
    target = "TARGET"
    datasource = DataFrameSource(multiimage_dataframe)
    schema = BitfountSchema()
    schema.add_datasource_tables(
        datasource,
        force_stypes={
            TABLE_NAME: {"image": ["image1", "image2"], "categorical": ["Date"]}
        },
        table_name=TABLE_NAME,
    )
    datasource.load_data()
    datasource.data = datasource.data.drop(
        columns=schema.get_feature_names(TABLE_NAME, SemanticType.TEXT)
    )
    datastructure = DataStructure(
        target=target, selected_cols=["image1", "image2", target], table=TABLE_NAME
    )
    datastructure.set_training_column_split_by_semantic_type(schema.tables[0])
    return _PyTorchDataset(
        datasource=datasource,
        target=target,
        selected_cols_semantic_types=datastructure.selected_cols_w_types,
        selected_cols=datastructure.selected_cols,
        schema=schema.get_table_schema(TABLE_NAME),
        data_split=DataSplit.TRAIN,
    )



"""Classes for dealing with Transformation Processing."""
from __future__ import annotations

import logging
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    List,
    Mapping,
    Optional,
    Set,
    Tuple,
    Type,
    Union,
    cast,
)

import numpy as np
import pandas as pd
from pandas.core.dtypes.common import is_numeric_dtype

from bitfount.data.types import DataSplit, SemanticType
from bitfount.transformations.base_transformation import (
    MultiColumnOutputTransformation,
    Transformation,
)
from bitfount.transformations.batch_operations import (
    AlbumentationsImageTransformation,
    BatchTimeOperation,
)
from bitfount.transformations.binary_operations import (
    AdditionTransformation,
    ComparisonTransformation,
    DivisionTransformation,
    MultiplicationTransformation,
    SubtractionTransformation,
)
from bitfount.transformations.dataset_operations import (
    CleanDataTransformation,
    NormalizeDataTransformation,
    ScalarAdditionDataTransformation,
    ScalarMultiplicationDataTransformation,
)
from bitfount.transformations.exceptions import (
    InvalidBatchTransformationError,
    MissingColumnReferenceError,
    NotColumnReferenceError,
    TransformationApplicationError,
)
from bitfount.transformations.references import _extract_col_ref
from bitfount.transformations.unary_operations import (
    InclusionTransformation,
    OneHotEncodingTransformation,
)

if TYPE_CHECKING:
    from bitfount.data.schema import TableSchema


logger = logging.getLogger(__name__)


class TransformationProcessor:
    """Processes Transformations on a given dataframe.

    :::caution

    The Transformation processor does not add any of the newly created columns
    to the Schema. This must be done separately after processing the transformations.

    :::

    Args:
        transformations: The list of transformations to apply.
        schema: The schema of the data to be transformed.
        col_refs: The set of columns referenced in those transformations.

    Attributes:
        transformations: The list of transformations to apply.
        schema: The schema of the data to be transformed.
        col_refs: The set of columns referenced in those transformations.
    """

    def __init__(
        self,
        transformations: List[Transformation],
        schema: Optional[TableSchema] = None,
        col_refs: Optional[Set[str]] = None,
    ):
        self.transformations = transformations
        self.col_refs = set() if col_refs is None else col_refs

        self.schema = schema
        if self.schema is not None:
            self._schema_cont_cols = self.schema.get_feature_names(
                SemanticType.CONTINUOUS
            )
            self._schema_cat_cols = self.schema.get_feature_names(
                SemanticType.CATEGORICAL
            )
        else:
            self._schema_cont_cols = []
            self._schema_cat_cols = []

        self._operators: Dict[Type[Transformation], Callable] = {
            AdditionTransformation: self._do_addition,
            CleanDataTransformation: self._do_clean_data,
            ComparisonTransformation: self._do_comparison,
            DivisionTransformation: self._do_division,
            InclusionTransformation: self._do_inclusion,
            MultiplicationTransformation: self._do_multiplication,
            NormalizeDataTransformation: self._do_normalize_data,
            OneHotEncodingTransformation: self._do_one_hot_encoding,
            SubtractionTransformation: self._do_subtraction,
            AlbumentationsImageTransformation: self._do_image_transformation,
            ScalarMultiplicationDataTransformation: self._do_scalar_multiplication,
            ScalarAdditionDataTransformation: self._do_scalar_addition,
        }
        cols_to_reverse: Mapping[str, Tuple[float, float]]

    def transform(self, data: pd.DataFrame) -> pd.DataFrame:
        """Performs `self.transformations` on `data` sequentially.

        Arguments to an operation are extracted by first checking if they are
        referencing another transformed column by checking for the name attribute.
        If not, we then check if they are referencing a non-transformed column by
        using a regular expression. Finally, if the regex comes back empty we take
        the argument 'as is' e.g. a string, integer, etc. After the transformations
        are complete, finally removes any columns that shouldn't be part of the final
        output.

        Args:
            data: The `pandas` dataframe to be transformed.

        Raises:
            MissingColumnReferenceError: If there is a reference to a non-existing
                column.
            TypeError: if there are clashes between column names or if unable
                to apply transformation.
        """
        data_columns = set(data.columns)
        # Check that all referenced columns are present in `data`
        missing_cols = sorted(self.col_refs.difference(data_columns))
        if missing_cols:
            raise MissingColumnReferenceError(
                [f"Reference to non-existent column: {c}" for c in missing_cols]
            )
        # Loop through transformations and perform them sequentially
        application_errors = []
        cols_to_drop = []
        for transformation in self.transformations:
            # Transformation name is always set
            transformation.name = cast(str, transformation.name)

            # Check transformation output doesn't clash with existing column
            if isinstance(transformation, MultiColumnOutputTransformation):
                clashes = data_columns.intersection(transformation.columns)
            else:
                clashes = data_columns.intersection([transformation.name])
            if clashes:
                application_errors.extend(
                    [
                        f"Output column {col_name}, "
                        f"from transformation {transformation.name}, "
                        f"clashes with an existing data column name."
                        for col_name in sorted(clashes)
                    ]
                )
                continue

            # Get operation as a function
            operation = self._operators[type(transformation)]

            # Add column(s) to `cols_to_drop` if it shouldn't end up in dataframe
            if not transformation.output:
                if isinstance(transformation, MultiColumnOutputTransformation):
                    cols_to_drop.extend(transformation.columns)
                else:  # normal transformation type
                    cols_to_drop.append(transformation.name)

            # Attempt to perform transformation. If there is a type mismatch
            # pandas will throw a type error which we catch and skip the
            # transformation.
            try:
                data = operation(data, transformation)
            except TypeError as e:
                application_errors.append(
                    f"Unable to apply transformation, skipping: "
                    f"{transformation.name}: {e}"
                )

        # Check all transformations were applied OK
        if application_errors:
            raise TransformationApplicationError(application_errors)

        # Drop any columns that should not be part of the final dataframe
        if cols_to_drop:
            return data.drop(cols_to_drop, axis=1)

        return data

    def batch_transform(self, data: np.ndarray, step: DataSplit) -> np.ndarray:
        """Performs batch transformations.

        Args:
            data: The data to be transformed at batch time as a numpy array.
            step: The step at which the data should be transformed.

        Raises:
            InvalidBatchTransformationError: If one of the specified transformations
                does not inherit from `BatchTimeOperation`.

        Returns:
            np.ndarray: The transformed data as a numpy array.
        """
        for transformation in self.transformations:
            # Check transformation is a batch time operation
            if not isinstance(transformation, BatchTimeOperation):
                raise InvalidBatchTransformationError(
                    f"{transformation._registry_name} not a batch time operation"
                )

            # Skip the transformation if it does not apply to the step
            if transformation.step != step:
                logger.debug(f"Skipping transformation: {transformation.name}")
                continue

            logger.debug(f"Applying transformation: {transformation.name}")
            # Get operation as a function
            operation = self._operators[type(transformation)]
            data = operation(data, transformation)

        return data

    @staticmethod
    def _apply_arg_conversion(
        data: pd.DataFrame, *args: Union[Transformation, str, Any]
    ) -> Union[Union[pd.Series, Any], List[Union[pd.Series, Any]]]:
        """Applies argument conversion to each of the supplied *args.

        Check, in order, if it is a:
            - Transformation instance (use transformation output column from data)
            - A column reference (use original column from data)
            - Other (use arg as is)

        Args:
            data: Data in a dataframe.
            *args: The args to convert.

        Returns: A single converted arg if only one was supplied, or a list of converted
            args in the same order they were supplied.
        """
        converted: List[Union[pd.Series, Any]] = []
        for arg in args:
            # See if it is a transformation
            if isinstance(arg, Transformation):
                converted.append(data[arg.name])
                continue

            # See if it is a column reference
            try:
                col_ref = _extract_col_ref(arg)
                converted.append(data[col_ref])
                continue
            except NotColumnReferenceError:
                pass

            # Otherwise, just use as is
            converted.append(arg)

        # If only one arg supplied, return single arg
        if len(converted) == 1:
            return converted[0]
        # Otherwise return list
        return converted

    @staticmethod
    def _get_list_of_col_refs(
        data: pd.DataFrame, cols: Union[str, List[str]], schema_cols: List[str]
    ) -> Tuple[List[str], bool]:
        """Gets the list of actual column references from a list of potentials.

        Generates a list of col names from a potential column list argument
        which could be:
            - a list of column references
            - "all": in which case use schema and data to generate list
            - a single column reference: wrap in a list

        Args:
            data:
                The data containing the columns.
            cols:
                The column reference(s) or "all".
            schema_cols:
                The list of schema columns to use to generate in case of "all".

        Returns:
            A tuple of the list of column references and a boolean indicating whether
            these have been "pre-extracted" (i.e. don't need to be compared to
            COLUMN_REFERENCE).
        """
        pre_extracted = False
        out_cols = cols

        if isinstance(out_cols, list):
            # If columns are not pre-extracted, they will have start
            # with "c:" or "col:"
            if not any(":" in col_name for col_name in out_cols):
                pre_extracted = True
            out_cols = out_cols
        elif isinstance(out_cols, str):
            if out_cols == "float":
                pre_extracted = True
                out_cols = cast(
                    list, data.select_dtypes(include=["float"]).columns.to_list()
                )
            elif out_cols == "all":
                # Extract all target columns from schema
                pre_extracted = True
                out_cols = [col for col in schema_cols if col in data.columns]
            else:
                # Wrap single column in iterable
                out_cols = [out_cols]
        return out_cols, pre_extracted

    @staticmethod
    def _do_addition(data: pd.DataFrame, t: AdditionTransformation) -> pd.DataFrame:
        """Performs addition transformation on `data` and returns it."""
        arg1, arg2 = TransformationProcessor._apply_arg_conversion(data, t.arg1, t.arg2)
        data[t.name] = arg1 + arg2
        return data

    def _do_clean_data(
        self, data: pd.DataFrame, t: CleanDataTransformation
    ) -> pd.DataFrame:
        """Cleans categorical and continuous columns in the data.

        Replaces infinities and NAs.
        """
        # Get columns
        cols, pre_extracted = self._get_list_of_col_refs(
            data, t.cols, self._schema_cat_cols + self._schema_cont_cols
        )
        if not pre_extracted:
            cols = [_extract_col_ref(col) for col in cols]

        # Clean columns
        for col_ref in cols:
            # categorical columns
            if col_ref in self._schema_cat_cols:
                data[col_ref] = data[col_ref].fillna("nan")
            # continuous columns
            elif col_ref in self._schema_cont_cols:
                data[col_ref] = data[col_ref].replace([np.inf, -np.inf], np.nan)
                data[col_ref] = data[col_ref].fillna(value=0.0)
            else:
                logger.warning(f"{col_ref} not found in Schema. Skipping cleaning.")

        return data

    @staticmethod
    def _do_comparison(data: pd.DataFrame, t: ComparisonTransformation) -> pd.DataFrame:
        """Performs comparison between arg1 and arg2 of comparison transformation."""
        arg1, arg2 = TransformationProcessor._apply_arg_conversion(data, t.arg1, t.arg2)

        arg1 = arg1.to_numpy()
        try:
            # If arg2 is a series-like
            arg2 = arg2.to_numpy()
        except AttributeError:
            # If arg2 is a constant
            arg2 = np.full_like(arg1, fill_value=arg2)
        conditions = [arg1 < arg2, arg1 == arg2, arg1 > arg2]
        choices = [-1, 0, 1]

        data[t.name] = np.select(conditions, choices, default=np.nan).tolist()

        return data

    @staticmethod
    def _do_division(data: pd.DataFrame, t: DivisionTransformation) -> pd.DataFrame:
        """Performs division transformation on `data` and returns it."""
        arg1, arg2 = TransformationProcessor._apply_arg_conversion(data, t.arg1, t.arg2)
        data[t.name] = arg1 / arg2
        return data

    @staticmethod
    def _do_inclusion(data: pd.DataFrame, t: InclusionTransformation) -> pd.DataFrame:
        # Only arg should be a column name in this transformation
        """Performs inclusion transformation on `data` and returns it."""
        arg = TransformationProcessor._apply_arg_conversion(data, t.arg)
        arg = cast(pd.Series, arg)
        data[t.name] = arg.str.contains(t.in_str)
        return data

    @staticmethod
    def _do_multiplication(
        data: pd.DataFrame, t: MultiplicationTransformation
    ) -> pd.DataFrame:
        """Performs multiplication transformation on `data` and returns it."""
        arg1, arg2 = TransformationProcessor._apply_arg_conversion(data, t.arg1, t.arg2)
        data[t.name] = arg1 * arg2
        return data

    def _do_normalize_data(
        self,
        data: pd.DataFrame,
        t: NormalizeDataTransformation,
    ) -> pd.DataFrame:
        """Normalizes numeric columns using their mean and stddev.

        Note: The default value of the transformation.cols for this
        transformation is "float". Results in mean of 0 and stddev of 1.
        """
        cols, pre_extracted = self._get_list_of_col_refs(
            data,
            t.cols,
            self._schema_cont_cols,
        )
        if not pre_extracted:
            cols = [_extract_col_ref(col) for col in cols]
        for col_ref in cols:
            if not is_numeric_dtype(data[col_ref]):
                raise TypeError(
                    f'Cannot normalize column "{col_ref}" as it is not numeric.'
                )
            mean = data[col_ref].mean()
            stddev = data[col_ref].std()
            data[col_ref] = (data[col_ref] - mean) / (1e-7 + stddev)
        if self.schema is not None:
            for col in cols:
                self.schema.features["continuous"][col].dtype = data[col].dtype

        return data

    def _do_scalar_multiplication(
        self, data: pd.DataFrame, t: ScalarMultiplicationDataTransformation
    ) -> pd.DataFrame:
        """Performs column multiplication with scalar.

        Only applicable to numeric columns.

        Raises:
            TypeError: if one of the referenced columns is not numeric.
            MissingColumnReferenceError: if one of the referenced columns is missing.
        """
        # If the scalar is a numerical value, we apply the scalar
        # multiplication to all numerical columns.
        if isinstance(t.scalar, (int, float)):
            cols, pre_extracted = self._get_list_of_col_refs(
                data, t.cols, self._schema_cont_cols
            )
            if not pre_extracted:
                cols = [_extract_col_ref(col) for col in cols]

            data[cols] = data[cols] * t.scalar
            return data
        # Otherwise, it is a dictionary mapping column names to scalar,
        # so we multiply each column with the scalar it is mapped to.
        else:
            missing_cols = set(t.scalar.keys()) - set(data.columns)
            if missing_cols:
                raise MissingColumnReferenceError(
                    [f"Reference to non-existent column: {c}" for c in missing_cols]
                )
            for col_ref, scalar in t.scalar.items():
                if not is_numeric_dtype(data[col_ref]):
                    raise TypeError(
                        f'Cannot normalize column "{col_ref}" as it is not numeric.'
                    )
                data[col_ref] = data[col_ref] * scalar
            return data

    def _do_scalar_addition(
        self, data: pd.DataFrame, t: ScalarAdditionDataTransformation
    ) -> pd.DataFrame:
        """Performs column addition with scalar.

        Only applicable to numeric columns.

        Raises:
            TypeError: if one of the referenced columns is not numeric.
            MissingColumnReferenceError: if one of the referenced columns is missing.
        """
        # If the scalar is a numerical value, we apply the scalar
        # addition to all numerical columns.
        if isinstance(t.scalar, (int, float)):
            cols, pre_extracted = self._get_list_of_col_refs(
                data, t.cols, self._schema_cont_cols
            )
            if not pre_extracted:
                cols = [_extract_col_ref(col) for col in cols]

            data[cols] = data[cols] + t.scalar
            return data
        # Otherwise, it is a dictionary mapping column names to scalar,
        # so we add the scalar to the column it is mapped to.
        else:
            missing_cols = set(t.scalar.keys()) - set(data.columns)
            if missing_cols:
                raise MissingColumnReferenceError(
                    [f"Reference to non-existent column: {c}" for c in missing_cols]
                )
            for col_ref, scalar in t.scalar.items():
                if not is_numeric_dtype(data[col_ref]):
                    raise TypeError(
                        f'Cannot normalize column "{col_ref}" as it is not numeric.'
                    )
                data[col_ref] = data[col_ref] + scalar
            return data

    @staticmethod
    def _do_one_hot_encoding(
        data: pd.DataFrame, t: OneHotEncodingTransformation
    ) -> pd.DataFrame:
        """Performs one hot encoding transformation on `data` and returns it."""
        # Get unencoded data and columns
        # Only one arg supplied so therefore is single series that's returned
        arg: pd.Series = cast(
            pd.Series, TransformationProcessor._apply_arg_conversion(data, t.arg)
        )
        ohe_cols: List[str] = sorted(t.columns)
        n_rows: int = len(arg)
        n_cols: int = len(ohe_cols)

        # Create an appropriately sized dataframe, filled with zeros
        ohe_df: pd.DataFrame = pd.DataFrame(
            data=np.zeros(shape=(n_rows, n_cols), dtype=np.int8), columns=ohe_cols
        )

        # First, set all values in the unknown column to 1 which correspond to
        # non-null values in the arg Series. These will be marked into the correct
        # column as we find matches, and if no match is found, they should be in
        # this column anyway.
        not_null_idxs = arg.index[arg.notnull()]
        ohe_df.loc[not_null_idxs, t.unknown_col] = 1

        # Iterate through the (value, target_col) pairs and set the correct column
        # to 1 for all locations the value is found. Sets the unknown column in
        # the same locations to 0.
        for val, target_col in t.values.items():
            match_idxs = arg.index[arg == val]
            ohe_df.loc[match_idxs, target_col] = 1
            ohe_df.loc[match_idxs, t.unknown_col] = 0

        return pd.concat([data, ohe_df], axis=1)

    @staticmethod
    def _do_subtraction(
        data: pd.DataFrame, t: SubtractionTransformation
    ) -> pd.DataFrame:
        """Performs subtraction transformation on `data` and returns it."""
        arg1, arg2 = TransformationProcessor._apply_arg_conversion(data, t.arg1, t.arg2)
        data[t.name] = arg1 - arg2
        return data

    @staticmethod
    def _do_image_transformation(
        data: np.ndarray, t: AlbumentationsImageTransformation
    ) -> np.ndarray:
        """Performs image transformation on `data` and returns it."""
        tfm = t.get_callable()
        return tfm(image=data)["image"]



"""Contains helper methods for transformation tests."""
from typing import Any, Type

import pytest

from bitfount.transformations.base_transformation import (
    TRANSFORMATION_REGISTRY,
    Transformation,
)
from bitfount.transformations.exceptions import TransformationRegistryError


def gen_name_fails(cls: Type[Transformation]) -> None:
    """Test that the _gen_name function fails for target Transformation."""
    with pytest.raises(TransformationRegistryError):
        cls._gen_name()


def gen_name_test(cls: Type[Transformation], registry_name: str, **kwargs: Any) -> None:
    """Test that the created cls instance has a correctly generated random name."""
    t = cls(**kwargs)
    assert t.name == f"{registry_name}_FAKE_HEX"


def registration_test(cls: Type[Transformation], registry_name: str) -> None:
    """Test that a class is correctly registered with the desired name."""
    assert registry_name in TRANSFORMATION_REGISTRY
    assert TRANSFORMATION_REGISTRY[registry_name] == cls



"""This package contains the core Bitfount codebase.

With the exception of `backends` which contains optional extra packages, every
subpackage or standalone module in this package must have an `__init__.py` file that
defines `__all__`.

Runners are explicitly excluded as they intended to be used by the scripts.
"""
#   ____  _ _    __                   _
#  | __ )(_) |_ / _| ___  _   _ _ __ | |_
#  |  _ \| | __| |_ / _ \| | | | '_ \| __|
#  | |_) | | |_|  _| (_) | |_| | | | | |_
#  |____/|_|\__|_|  \___/ \__,_|_| |_|\__|

import importlib as _importlib
import logging as _logging
import pkgutil as _pkgutil
import sys
from typing import Any
from typing import List as _List

from bitfount import (
    config,
    data,
    exceptions,
    federated,
    hub,
    metrics,
    models,
    storage,
    transformations,
    types,
    utils,
)
from bitfount.__version__ import __version__  # noqa: F401
from bitfount.config import *  # noqa: F403
from bitfount.data import *  # noqa: F403
from bitfount.exceptions import *  # noqa: F403
from bitfount.federated import *  # noqa: F403
from bitfount.hub import *  # noqa: F403
from bitfount.metrics import *  # noqa: F403
from bitfount.models import *  # noqa: F403
from bitfount.storage import *  # noqa: F403
from bitfount.transformations import *  # noqa: F403
from bitfount.types import *  # noqa: F403
from bitfount.utils import *  # noqa: F403

__all__: _List[str] = []

_logger = _logging.getLogger(__name__)

# Attempt to import backends if any exist
try:
    import bitfount.backends

    _backends_imported = True
except ModuleNotFoundError:
    _backends_imported = False

# If backends has been successfully imported, attempt to import each individual backend
# and add its __all__ to __all__
if _backends_imported:
    # Find all top-level subpackages in the backends package
    for _module_info in _pkgutil.iter_modules(
        bitfount.backends.__path__, f"{bitfount.backends.__name__}."
    ):
        _module = None
        # Attempt to import backend subpackage
        try:
            _module = _importlib.import_module(_module_info.name)
        except ImportError:
            pass

        # Add backend subpackage's __all__ to __all__
        if _module is not None:
            _imports: _List[str] = []
            try:
                _imports = _module.__dict__["__all__"]
                __all__.extend(_imports)
            except KeyError:
                _logger.error(f"Couldn't import {_module}: __all__ not defined.")

            # Add backend imports defined in __all__ to globals dictionary
            globals().update({k: getattr(_module, k) for k in _imports})


__all__.extend(config.__all__)
__all__.extend(data.__all__)
__all__.extend(exceptions.__all__)
__all__.extend(federated.__all__)
__all__.extend(hub.__all__)
__all__.extend(metrics.__all__)
__all__.extend(models.__all__)
__all__.extend(storage.__all__)
__all__.extend(transformations.__all__)
__all__.extend(types.__all__)
__all__.extend(utils.__all__)

# Currently, due to pdoc's reliance on `__all__`, we must iterate over `__all__`` to
# ignore every import in the documentation otherwise they become duplicated
# https://github.com/pdoc3/pdoc/issues/340
__pdoc__ = {}
for _obj in __all__:
    __pdoc__[_obj] = False

# Set stacktrace dependent on config
if not config.BITFOUNT_DEV_MODE:
    sys.tracebacklimit = config.BITFOUNT_TB_LIMIT

    # Jupyter/iPython has its own traceback system, so we need to handle that
    # differently
    try:
        # This function will be in scope if we're in jupyter/iPython
        ipython = get_ipython()  # type: ignore[name-defined] # Reason: see comment # noqa: B950, F405

        def _hide_traceback(
            exc_tuple: Any = None,
            filename: Any = None,
            tb_offset: Any = None,
            exception_only: Any = False,
            running_compiled_code: Any = False,
        ) -> Any:
            etype, value, tb = sys.exc_info()
            return ipython._showtraceback(
                etype, value, ipython.InteractiveTB.get_exception_only(etype, value)
            )

        ipython.showtraceback = _hide_traceback
    except NameError:
        # We are not in a Jupyter/iPython environment
        pass



"""Tests for PyTorch-specific utils."""
from typing import Callable

from _pytest.logging import LogCaptureFixture
from _pytest.monkeypatch import MonkeyPatch

from bitfount.backends.pytorch.utils import _autodetect_gpu
import bitfount.config
from bitfount.data.datafactory import _get_default_data_factory
from tests.utils.helper import backend_test, unit_test


def mock_device_count(count: int = 0) -> Callable[[], int]:
    """Mock device counter for CUDA."""

    def f() -> int:
        return count

    return f


@backend_test
@unit_test
class TestUtils:
    """Tests for util functions."""

    def test_autodetect_gpu_cpu_only(
        self, caplog: LogCaptureFixture, monkeypatch: MonkeyPatch
    ) -> None:
        """Tests auto-detecting GPU count when only CPU."""
        # Mock out CUDA device count
        caplog.set_level("INFO")
        monkeypatch.setattr("torch.cuda.device_count", mock_device_count(0))

        count = _autodetect_gpu()

        assert count == 0
        assert (
            caplog.records[0].msg == "No supported GPU detected. Running model on CPU."
        )

    def test_autodetect_1_gpu(
        self, caplog: LogCaptureFixture, monkeypatch: MonkeyPatch
    ) -> None:
        """Tests auto-detecting GPU when only one GPU."""
        # Mock out CUDA device count
        caplog.set_level("INFO")
        monkeypatch.setattr("torch.cuda.device_count", mock_device_count(1))
        monkeypatch.setattr("torch.cuda.get_device_name", lambda x: f"GPU_{x}")

        count = _autodetect_gpu()

        assert count == 1
        assert (
            caplog.records[0].msg == "CUDA support detected. GPU (GPU_0) will be used."
        )

    def test_autodetect_multiple_gpu(
        self, caplog: LogCaptureFixture, monkeypatch: MonkeyPatch
    ) -> None:
        """Tests auto-detecting GPU when multiple GPUs."""
        # Mock out CUDA device count
        caplog.set_level("INFO")
        monkeypatch.setattr("torch.cuda.device_count", mock_device_count(2))
        monkeypatch.setattr("torch.cuda.get_device_name", lambda x: f"GPU_{x}")

        count = _autodetect_gpu()

        assert count == 1
        assert caplog.records[0].levelname == "WARNING"
        assert (
            caplog.records[0].msg
            == "Bitfount model currently only supports one GPU. Will use GPU 0 (GPU_0)."
        )
        assert (
            caplog.records[1].msg == "CUDA support detected. GPU (GPU_0) will be used."
        )


@backend_test
@unit_test
class TestDefaultDataFactoryLoading:
    """Tests for loading the default data factory when PyTorch installed."""

    def test_load_pytorch_default_data_factory(self, monkeypatch: MonkeyPatch) -> None:
        """Test that the default data factory can load."""
        # Ensure PyTorch is set as the engine variable
        monkeypatch.setattr(
            "bitfount.config.BITFOUNT_ENGINE", bitfount.config._PYTORCH_ENGINE
        )

        # Create a fake class and set that as the PyTorch data factory
        class FakeDataFactory:
            pass

        monkeypatch.setattr(
            "bitfount.backends.pytorch.data.datafactory._PyTorchDataFactory",
            FakeDataFactory,
        )

        df = _get_default_data_factory()
        assert isinstance(df, FakeDataFactory)



"""Main Components for our pod runner UI."""
import logging
import sys
from typing import Any

# Only installed on some platforms
# mypy_reason: Only installed on some platforms, only used here, and heavy C code make
#       this a poor candidate for stub generation. Easier to ignore.
import wx  # type: ignore[import]  # Reason: see above
import wxasync  # type: ignore[import] # Reason: see above

from bitfount.__version__ import __version__
from bitfount.runners.pod_runner import setup_pod_from_config_file

logger = logging.getLogger("bitfount")


class WxTextCtrlHandler(logging.Handler):
    """Loging handler for showing logs in the UI."""

    def __init__(self, ctrl: Any):
        logging.Handler.__init__(self)
        self.ctrl = ctrl

    def emit(self, record: logging.LogRecord) -> None:
        """Emits and formats a record."""
        s = self.format(record) + "\n"
        wx.CallAfter(self.ctrl.WriteText, s)


class Frame(wx.Frame):
    """Main Class for the pod runner UI."""

    def __init__(self) -> None:
        TITLE = "Bitfount Pod Runner"
        wx.Frame.__init__(self, None, wx.ID_ANY, TITLE, size=(500, 400))

        # Create the panel.
        self.panel = wx.Panel(self, wx.ID_ANY)

        # Create the text box that will host the logs.
        log = wx.TextCtrl(
            self.panel,
            wx.ID_ANY,
            size=(200, 2000),
            style=wx.TE_MULTILINE | wx.TE_READONLY | wx.HSCROLL,
        )

        # Create linebreaks (used for styling purposes only).
        linebreak = wx.StaticText(self.panel, label="")
        linebreak2 = wx.StaticText(self.panel, label="")

        # Create the button for the FileOpen Dialog box.
        self.load_btn = wx.Button(
            self.panel, wx.ID_ANY, "Select pod configuration file"
        )
        # Placeholder for user-provided filename.
        self.NewText = wx.StaticText(self.panel, label="")
        self.NewText.Hide()

        # Create `Start pod` and `Exit` buttons.
        self.start_btn = wx.Button(self.panel, wx.ID_ANY, "Start pod", size=(120, 20))
        self.stop_btn = wx.Button(self.panel, wx.ID_ANY, "Exit", size=(120, 20))

        # Disable `Start pod` until user specifies a file.
        self.start_btn.Disable()

        # Bind the created buttons to event functions.
        wxasync.AsyncBind(wx.EVT_BUTTON, self.onStartButton, self.start_btn)
        self.Bind(wx.EVT_BUTTON, self.onLoadButton, self.load_btn)
        self.Bind(wx.EVT_BUTTON, self.onStopButton, self.stop_btn)

        # Create the main sizer panel that will host everything.
        self.main_sizer = wx.BoxSizer(wx.VERTICAL)

        # Create the sizer for the file chooser and filename.
        file_sizer = wx.BoxSizer(wx.VERTICAL)
        file_sizer.Add(linebreak, 0, wx.ALL | wx.CENTER, 5)
        file_sizer.Add(self.load_btn, 0, wx.ALL | wx.CENTER, 5)
        file_sizer.Add(self.NewText, flag=wx.CENTER, border=5)
        file_sizer.Add(linebreak2, 0, wx.ALL | wx.CENTER, 5)

        # Create the sizer for the buttons.
        self.btn_sizer = wx.BoxSizer(wx.HORIZONTAL)
        self.btn_sizer.Add(self.start_btn, 0, wx.ALL | wx.LEFT, 5)
        self.btn_sizer.Add(self.stop_btn, 0, wx.ALL | wx.RIGHT, 5)

        # Add the sizers and the log text box to our main sizer.
        self.main_sizer.Add(file_sizer, 0, wx.ALL | wx.CENTER)
        self.main_sizer.Add(self.btn_sizer, 0, wx.ALL | wx.CENTER)
        self.main_sizer.Add(log, 1, wx.ALL | wx.EXPAND, 5)
        self.panel.SetSizer(self.main_sizer)

        # Add the new log handler to our logger.
        handler = WxTextCtrlHandler(log)
        logger.addHandler(handler)

        # Log the running bitfount version.
        logger.log(logging.INFO, f"Running Bitfount version {__version__}")
        FORMAT = "%(asctime)s %(levelname)s %(message)s"
        handler.setFormatter(logging.Formatter(FORMAT))

        logger.setLevel(logging.INFO)

    def onLoadButton(self, event: Any) -> None:
        """Event config for the load file button."""
        cfgFrame = wx.Frame(None, -1)
        dlg = wx.FileDialog(
            cfgFrame,
            "Choose Pod config file.",
            "",
            "",
            "YAML Files (*.yaml; *.yml)|*.yaml; *.yml",
            wx.FD_OPEN | wx.FD_FILE_MUST_EXIST,
        )

        # Check that the user did not cancel the file selection.
        if dlg.ShowModal() != wx.ID_CANCEL:
            # Get filename path and display it.
            self.filename = dlg.GetPath()
            self.NewText.SetLabel(self.filename)
            self.NewText.Show()
            # Refresh and update the text field.
            self.NewText.Refresh()
            self.NewText.Update()
            # Update layout.
            self.main_sizer.Layout()
            # Enable `Start pod` button once a file is chosen.
            self.start_btn.Enable()
            self.panel.Refresh()
            self.panel.Update()

    async def onStartButton(self, event: Any) -> None:
        """Event config for the `Start pod` button."""
        # Disable the button once it is pressed.
        self.start_btn.Disable()
        # Run pod and send logs.
        try:
            pod = setup_pod_from_config_file(self.filename)
            # Change the label for the `Exit` button once pod starts.
            self.stop_btn.SetLabel("Stop pod and exit")
            self.btn_sizer.Layout()
            await pod.start_async()

        except Exception as e:
            # Catch Exceptions and log them.
            logger.log(logging.ERROR, e)

    def onStopButton(self, event: Any) -> None:
        """Event config for the `Start pod` button.

        This button will just stop pod (if any pod is running) and exit the app.
        """
        # TODO: [BIT-299] Add pod shutdown signal here.
        sys.exit(0)



#!/usr/bin/env python3
"""Script to run a Pod that acts as a compute and data provider."""
from asyncio import get_event_loop
import logging
from os import PathLike
import platform
import sys
from typing import Union

import fire

from bitfount import config
from bitfount.runners.pod_runner import setup_pod_from_config_file
from bitfount.runners.utils import setup_loggers

loggers = setup_loggers([logging.getLogger("bitfount")])

config._BITFOUNT_CLI_MODE = True


def run(path_to_config_yaml: Union[str, PathLike]) -> None:
    """Runs a pod from a config file."""
    pod = setup_pod_from_config_file(path_to_config_yaml)
    pod.start()


def main() -> None:
    """Script entry point."""
    # If an argument has been provided to the script, we will run in headless mode.
    # The first element in sys.argv is always the script name so to check if any
    # arguments have been provided we check if the length of sys.argv is greater than 1.
    if len(sys.argv) > 1:
        fire.Fire(run)

    # Otherwise we will run in interactive mode (depending on the OS).
    elif platform.system() == "Windows":
        import easygui as easygui  # type: ignore[import] # Reason: Only present on windows. # noqa: B950

        # Windows doesn't work well with wxasync, so using easygui instead of wx.
        filename = easygui.fileopenbox(filetypes=[["*.yaml", "*.yml", "YAML files"]])
        if filename:
            try:
                run(filename)
            except Exception as e:
                print(e)
                sys.exit(1)
        else:
            print("No config file provided.")

    elif sys.platform == "darwin":

        # Skipping analyzing "wxasync": module is installed,
        # but missing library stubs or py.typed marker
        # This is only used for Mac but must be imported here to avoid mypy errors on
        # non-mac platforms.
        # mypy_reason: Only installed on some platforms, only used here, and heavy
        #              C code make this a poor candidate for stub generation. Easier
        #              to ignore.
        import wxasync  # type: ignore[import] # Reason: see above

        from scripts.pod_ui_utils import Frame

        # Use wxPython if client is not headless
        app = wxasync.WxAsyncApp(redirect=True)
        try:
            Frame().Show()
            loop = get_event_loop()
            loop.run_until_complete(app.MainLoop())

        except Exception as e:
            print(e)
            print("The file provided does not match the required format.")
            sys.exit(1)

    else:
        print(
            f"{sys.platform} OS does not support a GUI. Please run in headless mode "
            f"and provide a config file.",
            file=sys.stderr,
        )


if __name__ == "__main__":
    main()



"""Module for setting up the pod keys."""
from dataclasses import dataclass
from pathlib import Path
from typing import cast

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives._serialization import Encoding
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey, RSAPublicKey
from cryptography.hazmat.primitives.serialization import PublicFormat

from bitfount.federated.encryption import _RSAEncryption
from bitfount.federated.logging import _get_federated_logger

PRIVATE_KEY_FILE = "pod_rsa.pem"
PUBLIC_KEY_FILE = "pod_rsa.pub.pem"


logger = _get_federated_logger(__name__)


@dataclass
class PodKeys:
    """A public-private key pair for a `Pod`.

    Args:
        public_key: The public key.
        private_key: The private key.
    """

    public: RSAPublicKey
    private: RSAPrivateKey


def _generate_key_pair(private_key_path: Path, public_key_path: Path) -> PodKeys:
    """Generates, saves and returns an RSA key pair.

    Args:
        private_key_path: the path to save the private key to
        public_key_path: the path to save the public key to

    Returns:
        A tuple of (private key, public key)
    """
    private_key_path.parent.mkdir(exist_ok=True, parents=True)
    public_key_path.parent.mkdir(exist_ok=True, parents=True)

    private_key, public_key = _RSAEncryption.generate_key_pair()

    private_key_path.write_bytes(_RSAEncryption.serialize_private_key(private_key))
    public_key_path.write_bytes(
        public_key.public_bytes(Encoding.OpenSSH, PublicFormat.OpenSSH)
    )
    return PodKeys(private=private_key, public=public_key)


def _load_key_pair(private_key_path: Path, public_key_path: Path) -> PodKeys:
    """Loads an existing RSA key pair.

    Args:
        private_key_path: the path to load the private key from
        public_key_path: the path to load the public key from

    Returns:
        A tuple of (private key, public key)
    """
    bytes_ = public_key_path.read_bytes()
    public_key = serialization.load_ssh_public_key(bytes_, default_backend())
    private_key = serialization.load_pem_private_key(
        private_key_path.read_bytes(), password=None, backend=default_backend()
    )

    # Reassure mypy
    public_key = cast(RSAPublicKey, public_key)
    private_key = cast(RSAPrivateKey, private_key)

    return PodKeys(public=public_key, private=private_key)


def _get_pod_keys(key_directory: Path) -> PodKeys:
    """Gets the encryption key pair for the pod.

    Get the pod keys from the target directory, generating them if they don't
    already exist.

    Args:
        key_directory: the directory where the keys are located or should be saved

    Returns:
        A tuple of (private key, public key)
    """
    private_key_path = key_directory / PRIVATE_KEY_FILE
    public_key_path = key_directory / PUBLIC_KEY_FILE
    if private_key_path.exists() and public_key_path.exists():
        pod_keys = _load_key_pair(private_key_path, public_key_path)
    else:
        pod_keys = _generate_key_pair(private_key_path, public_key_path)
    return pod_keys
